# megamicros.antenna.py
#
# ® Copyright 2024-2025 Bimea
# Author: bruno.gas@bimea.io
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Core module of the Megamicros systems library

Features:
    - Base class for microphone arrays modelling.
    - Management of available and active microphones and analog inputs
    - Configuration of sampling frequency, data type, frame length, and acquisition duration
    - Support for counter activation
    - H5 local recording flag

Examples:
    Basic usage::

        from megamicros import log
        from megamicros.core.mu import Megamicros

        antenna = Megamicros()

        print( "Available MEMS: ", antenna.available_mems )
        print( "Available Analogs: ", antenna.available_analogs )

        antenna.run(
            mems=antenna.available_mems,
            sampling_frequency=50000,
            duration=10,
            frame_length=1024,
            datatype='int32'
        )
        antenna.wait()

        # Print frames stored in the queue
        print(f"queue content : {antenna.queue_content} frames")

        # Retrieve data from the queue
        for data in antenna:
            print( f"data={data}" )

    Advanced usage::

        See the Notebooks for advanced usage examples.

Documentation:
    Full MegaMicros documentation is available at: https://readthedoc.bimea.io
"""


from ctypes import addressof, byref, sizeof, create_string_buffer
import time
import platform
import enum
import queue
import numpy as np

from ..usb import Usb
from ..log import log
from ..exception import MuException
from .base import MemsArray, TRANSFER_DATAWORDS_SIZE

# MegaMicro hardware commands
MU_CMD_RESET					= b'\x00'									# Reset: power off the microphones
MU_CMD_INIT						= b'\x01'									# Sampling frequency setting
MU_CMD_START					= b'\x02'									# Acquisition running command
MU_CMD_STOP						= b'\x03'									# Acquisition stopping command
MU_CMD_COUNT					= b'\x04'									# Number of expected samples for next acquisition running
MU_CMD_ACTIVE					= b'\x05'									# Channels selection (MEMs, analogics, counter and status activating)
MU_CMD_PURGE					= b'\x06'									# Purge FiFo. No doc found about this command
MU_CMD_DELAY					= b'\x07'									# Test and tunning command. Not used in production mode. See documentation (no write function provided so far)
MU_CMD_DATATYPE					= b'\x09'									# Set datatype
MU_CMD_FX3_RESET				= 0xC0										# Init FX3 usb controler
MU_CMD_FX3_PH					= 0xC4										# External FPGA reset (hard reset)
MU_CMD_FPGA_0					= 0xB0										# Send a 0 byte command to FPGA
MU_CMD_FPGA_1					= 0xB1										# Send a 1 byte command to FPGA
MU_CMD_FPGA_2					= 0xB2										# Send a 2 byte command to FPGA
MU_CMD_FPGA_3					= 0xB3										# Send a 3 byte command to FPGA
MU_CMD_FPGA_4					= 0xB4										# Send a 4 byte command to FPGA

# Memgamicros hardware code values																	
MU_CODE_DATATYPE_INT32			= b'\x00'									# Int32 datatype code
MU_CODE_DATATYPE_FLOAT32		= b'\x01'									# Float32 datatype code

# MegaMicro receiver properties
MU_BEAM_MEMS_NUMBER				= 8											# MEMS number per beam
MU_MEMS_UQUANTIZATION			= 24										# MEMs unsigned quantization 
MU_MEMS_QUANTIZATION			= MU_MEMS_UQUANTIZATION - 1					# MEMs signed quantization 
MU_MEMS_AMPLITUDE				= 2**MU_MEMS_QUANTIZATION					# MEMs maximal amlitude value for "int32" data type
MU_MEMS_SENSIBILITY				= 3.54e-6 	                                # MEMs sensibility factor (-26dBFS for 104 dB that is 3.17 Pa)
MU_DEFAULT_DATATYPE             = 'int32'                                   # Datatype for FPGA megamicros data 

# Default run propertie's values
DEFAULT_TIME_ACTIVATION			= 200										# Waiting time after MEMs powering in milliseconds
DEFAULT_TIME_ACTIVATION_RESET	= 10										# Waiting time between commands of the MegaMicro device reset sequence in milliseconds
DEFAULT_CLOCKDIV				= 0x09										# Default internal acquisition clock value
DEFAULT_SELFTEST_DURATION       = 0.1                                       # Default selftest duration in seconds     
DEFAULT_START_TRIGG_STATUS      = False								        # Default start trigger status (external hard (True) or internal soft (False))
DEFAULT_MEMS_SENSIBILITY        = 3.54e-6                                   # Default MEMS sensitivity (racine(2)/400 000 = 3,54µPa/digit)
DEFAULT_ANALOGS_SENSIBILITY     = 0.33                                      # Default analogs sensibility in V/digit (0.33 V/digit: 0x00FFFFFF on 24bits <-> 5.65Vcc)
DEFAULT_CLOCK_DIVIDER_REFERENCE = 500000                                    # Constant for the clock divider (500kHz)
AIKHOUS_CLOCK_DIVIDER_REFERENCE = 480000                                    # Constant for the clock divider for Aikhous systems (480kHz)
DEFAULT_SYNC_DELAY              = 10                                        # Default synchronization delay (10 for usual systems, 8 for Aikhous systems)
AIKHOUS_SYNC_DELAY              = 8                                         # Default synchronization delay for Aikhous systems

CONTROL_DATA_FAILURE            = False                                     # Perform control data failure if True
EXIT_ON_DATA_FAILURE            = True                                      # Exit on data failure (when data are lost during transfer)

MU_BUS_ADDRESS                  = 0x00                                      # Default USB bus address for MegaMicros devices
USB_DEFAULT_WRITE_TIMEOUT       = 1000                                      # Default USB write timeout in milliseconds
USB_DEFAULT_TRANSFER_TIMEOUT    = 1000                                      # Default USB transfer timeout in milliseconds
USB_DEFAULT_BUFFERS_NUMBER      = 8                                         # Default USB buffers number for data acquisition
USB_DEFAULT_QUEUE_LENGTH        = 0                                         # Default USB transfer queue length in number of frames (0 means infinite queueing)
USB_DEFAULT_QUEUE_TIMEOUT       = 2                                         # Default USB transfer queue get timeout in seconds (delay until the queue is considered as empty)


class Megamicros(MemsArray):
    """
    class MegaMicros
    
    Main class to handle MegaMicros devices. Support 32, 256 and 1024 systems
    """

    Systems = {
        # no megamicros type specified
        'unknown': {
            'name': 'unknown',
            'vendor_id': 0x0000,
            'product_id': 0x0000,
            'endpoint_in': 0x00,
            'bus_address': 0x00,
            'beams': 0,
            'analogs': 0,
            'counters': 0,
            'status': 0
        },
        # 32 channels megamicros device with USB2 port
        'mu32usb2': {
            'name': 'mu32usb2',
            'vendor_id': 0xFE27,
            'product_id': 0xAC00,
            'endpoint_in': 0x81,
            'bus_address': 0x00,
            'beams': 4,
            'analogs': 0,
            'counters': 1,
            'status': 1

        },
        # 32 channels megamicros device with USB3 port
        'mu32': {
            'name': 'mu32',
            'vendor_id': 0xFE27,
            'product_id': 0xAC03,
            'endpoint_in': 0x81,
            'bus_address': 0x00,
            'beams': 4,
            'analogs': 0,
            'counters': 1,
            'status': 1
        },
        # 32 channels megamicros device with USB3 port and one or more analogs
        'mu32a': {
            'name': 'mu32a',
            'vendor_id': 0xFE27,
            'product_id': 0xAC03,
            'endpoint_in': 0x81,
            'bus_address': 0x00,
            'beams': 4,
            'analogs': 2,
            'counters': 1,
            'status': 1
        },
        # 128 channels megamicros device with USB2 port (deprecated)
        'mu128': {
            'name': 'mu128',
            'vendor_id': 0xFE27,
            'product_id': 0x0000,
            'endpoint_in': 0x81,
            'bus_address': 0x00,
            'beams': 16,
            'analogs': 0,
            'counters': 1,
            'status': 1
        },
        # 256 channels megamicros device with USB3 port
        'mu256': {
            'name': 'mu256',
            'vendor_id': 0xFE27,
            'product_id': 0xAC02,
            'endpoint_in': 0x81,
            'bus_address': 0x00,
            'beams': 32,
            'analogs': 4,
            'counters': 1,
            'status': 1
        },
        # 256 channels megamicros Aikhous device with USB3 port
        'mu256h': {
            'name': 'mu256h',
            'vendor_id': 0xFE27,
            'product_id': 0xAC02,
            'endpoint_in': 0x81,
            'bus_address': 0x00,
            'beams': 32,
            'analogs': 4,
            'counters': 1,
            'status': 1
        },
        # 1024 channels megamicros device with USB3 port
        'mu1024': {
            'name': 'mu1024',
            'vendor_id': 0xFE27,
            'product_id': 0xAC01,
            'endpoint_in': 0x81,
            'bus_address': 0x00,
            'beams': 128,
            'analogs': 16,
            'counters': 4,
            'status': 4
        },
    }

    def __init__(self):
        """
        @brief Constructor
        """
        super().__init__()
        self.__system: dict = Megamicros.Systems['unknown']                 # Megamicros system type
        self.__usb: Usb=Usb()                                               # USB interface instance
        self.__usb_buffers_number: int=USB_DEFAULT_BUFFERS_NUMBER           # USB buffers number used in USB data acquisition process
        self.__available_analogs: list[int]=[]                              # Available analogs (connected and ok on the antenna)
        self.__analogs: list[int]=[]                                        # Activated analogs
        self.__analogs_sensibility: float=DEFAULT_ANALOGS_SENSIBILITY       # Analogs sensibility in V/digit (default to 0.33 V/digit: 0x00FFFFFF on 24bits <-> 5.65Vcc)
        self.__counters: list[int]=[0]                                      # Activated counters
        self.__status: list[int]=[]                                         # Activated status channels
        self.__clock_divider_reference: int=DEFAULT_CLOCK_DIVIDER_REFERENCE # Clock divider reference (sr = cdr/(clockdiv+1), default to 500kHz for usual systems, 480kHz for Aikhous systems) 
        self.__clockdiv: int=9                                              # Clock divider (default to 9 for 50kHz or 48kHz sampling frequencies)
        self.__sync_delay: int=10                                           # Default synchronization delay (10 for usual systems, 8 for Aikhous systems)
        self.__start_trigg_status: bool=False                               # Start trigger status (external hard (True) or internal soft (False))
        self.__time_activation: int=DEFAULT_TIME_ACTIVATION                 # Waiting time between commands of the MegaMicro device reset sequence in milliseconds

        # Check connected devices and open connection to the first found
        self.checkAndOpenDevice()
        mems_power, analogs_power = self.selftest( DEFAULT_SELFTEST_DURATION )

        # Set default sampling frequency in accordance with default megamicros systems
        self.setClockdiv( DEFAULT_CLOCKDIV )

        log.info(" .Megamicros device initialized")

    def __del__( self ):
        log.info( ' .Megamicros resources cleaned up' )

    @property
    def usb( self ) -> Usb:
        return self.__usb

    @property
    def analogs( self ) -> int:
        return self.__analogs
    
    @property
    def available_analogs( self ) -> float:
        return self.__available_analogs

    @property
    def analogs_sensibility( self ) -> float:
        return self.__analogs_sensibility

    @property
    def counters( self ) -> list[int]:
        """ Return the list of activated counters"""
        return self.__counters
    
    @property
    def counter( self ) -> bool:
        """ Return True if at least counter 0 is activated
        """
        return 0 in self.__counters
    
    @property
    def status(self) -> list[int]:
        """ Return the list of activated status channels """
        return self.__status

    @property
    def channels_number( self ) -> int:
        """ Return the total number of channels (MEMS, analogs, counters, status) """
        return len(self.mems) + len(self.analogs) + len(self.counters) + len(self.status)

    @property
    def clockdiv( self ) -> int:
        """ Return the clock divider (default to 9 for 50kHz or 48kHz sampling frequencies) """
        return self.__clockdiv
    
    @property
    def sync_delay( self ) -> int:
        """ Return the synchronization delay (10 for usual systems, 8 for Aikhous systems) """
        return self.__sync_delay

    @property
    def queue( self ) -> Usb.Queue:
        """ Get the USB queue instance """
        return self.__usb.queue

    @property
    def queue_size( self ) -> int:
        """ Get the USB queue size (e.g. maxsize) in bytes"""
        return self.__usb.queue.maxsize

    @property
    def queue_length( self ) -> int:
        """ Get the USB queue length in number of frames """
        if self.channels_number == 0 or self.frame_length == 0:
            return 0
        else:
            return self.__usb.queue.maxsize // ( self.frame_length * self.channels_number * TRANSFER_DATAWORDS_SIZE )

    @property
    def queue_content( self ) -> int:
        """ Get the number of elements currently in the USB queue """
        return self.__usb.queue.qsize()

    @property
    def transfer_lost( self ) -> int:
        return self.usb.transfer_lost

    @property
    def time_activation( self ) -> int:
        return self.__time_activation

    @property
    def insfos( self ) -> dict:
        """ Get current Megamicros configuration as a dictionary """
        infos_dict = {
            'system_name': self.__system['name'],
            'available_mems': self.available_mems,
            'active_mems': self.mems,
            'mems_sensibility': self.mems_sensibility,
            'available_analogs': self.available_analogs,
            'active_analogs': self.analogs,
            'analogs_sensibility': self.analogs_sensibility,
            'counters': self.counters,
            'status': self.status,
            'clockdiv': self.clockdiv,
            'sampling_frequency': self.sampling_frequency,
            'sync_delay': self.sync_delay,
            'datatype': self.datatype,
            'duration': self.duration,
            'frame_length': self.frame_length,
            'frame_duration': self.frame_duration,
            'channels_number': self.channels_number,
            'counter': self.counter,
            'h5_recording': self.__h5_recording
        }
        return infos_dict

    def setAvailableMems( self, mems: list[int] ) -> None:
        """ Set the available MEMs (connected and ok on the antenna)

        Parameters
        ----------
        mems: list[int]
            The available MEMs list
        """
        mems_number = self.__system['beams'] * MU_BEAM_MEMS_NUMBER
        # Check if a MEMS number in the list is >= mems_number:
        for m in mems:
            if m >= mems_number:
                raise MuException( f"Megamicros system {self.__system['name']} has only {mems_number} MEMS, cannot set available MEMS {mems}" )

        super().setAvailableMems(mems)

    def setAvailableAnalogs( self, analogs: list[int] ) -> None:
        """ Set the available analogs (connected and ok on the antenna)

        Parameters
        ----------
        analogs: list[int]
            The available analogs list
        """
        analogs_number = self.__system['analogs']
        # Check if an analog number in the list is >= analogs_number:
        for a in analogs:
            if a >= analogs_number:
                raise MuException( f"Megamicros system {self.__system['name']} has only {analogs_number} analogs, cannot set available analogs {analogs}" )
        self.__available_analogs = analogs

    def setActiveAnalogs( self, analogs: tuple ) -> None :
        """ Activate analogs
        
        Parameters:
        -----------
        analogs : tuple
            list or tuple of analogs number to activate
        """

        self.__analogs = analogs

    def setSyncDelay( self, delay: int ) -> None:
        """ Set the synchronization delay (10 for usual systems, 8 for Aikhous systems)

        Parameters:
        -----------
        delay : int
            The synchronization delay value
        """
        self.__sync_delay = delay

    def setQueueLength( self, length: int ) -> None:
        """ Set USB transfer queue length in number of frames
        """
        self.usb.setQueueSize( length * self.frame_length * self.channels_number * TRANSFER_DATAWORDS_SIZE )

    def checkAndOpenDevice(self) -> None:
        """
        Check and open the MegaMicros USB device
        
        Throw MuException in case of error during the USB transfer
        """
        try:
            if self.__usb.isOpened():
                raise MuException("MegaMicros device is already opened")

            if self.__usb.checkDeviceByVendorProduct(Megamicros.Systems['mu32usb2']['vendor_id'], Megamicros.Systems['mu32usb2']['product_id']):
                self.__system = Megamicros.Systems['mu32usb2']
                self.__clock_divider_reference = DEFAULT_CLOCK_DIVIDER_REFERENCE
                self.__sync_delay = DEFAULT_SYNC_DELAY
                self.__usb.open(
                    Megamicros.Systems['mu32usb2']['vendor_id'],
                    Megamicros.Systems['mu32usb2']['product_id'],
                    Megamicros.Systems['mu32usb2']['bus_address'],
                    Megamicros.Systems['mu32usb2']['endpoint_in']
                )

            if self.__usb.checkDeviceByVendorProduct(Megamicros.Systems['mu32']['vendor_id'], Megamicros.Systems['mu32']['product_id']):
                self.__system = Megamicros.Systems['mu32']
                self.__clock_divider_reference = DEFAULT_CLOCK_DIVIDER_REFERENCE
                self.__sync_delay = DEFAULT_SYNC_DELAY
                self.__usb.open(
                    Megamicros.Systems['mu32']['vendor_id'],
                    Megamicros.Systems['mu32']['product_id'],
                    Megamicros.Systems['mu32']['bus_address'],
                    Megamicros.Systems['mu32']['endpoint_in']
                )

            elif self.__usb.checkDeviceByVendorProduct(Megamicros.Systems['mu128']['vendor_id'], Megamicros.Systems['mu128']['product_id']):
                self.__system = Megamicros.Systems['mu128']
                self.__clock_divider_reference = DEFAULT_CLOCK_DIVIDER_REFERENCE
                self.__sync_delay = DEFAULT_SYNC_DELAY
                self.__usb.open(
                    Megamicros.Systems['mu128']['vendor_id'],
                    Megamicros.Systems['mu128']['product_id'],
                    Megamicros.Systems['mu128']['bus_address'],
                    Megamicros.Systems['mu128']['endpoint_in']
                )

            elif self.__usb.checkDeviceByVendorProduct(Megamicros.Systems['mu256']['vendor_id'], Megamicros.Systems['mu256']['product_id']):
                self.__system = Megamicros.Systems['mu256']
                self.__clock_divider_reference = DEFAULT_CLOCK_DIVIDER_REFERENCE
                self.__sync_delay = DEFAULT_SYNC_DELAY
                self.__usb.open(
                    Megamicros.Systems['mu256']['vendor_id'],
                    Megamicros.Systems['mu256']['product_id'],
                    Megamicros.Systems['mu256']['bus_address'],
                    Megamicros.Systems['mu256']['endpoint_in']
                )

            elif self.__usb.checkDeviceByVendorProduct(Megamicros.Systems['mu1024']['vendor_id'], Megamicros.Systems['mu1024']['product_id']):
                self.__system = Megamicros.Systems['mu1024']
                self.__clock_divider_reference = DEFAULT_CLOCK_DIVIDER_REFERENCE
                self.__sync_delay = DEFAULT_SYNC_DELAY
                self.__usb.open(
                    Megamicros.Systems['mu1024']['vendor_id'],
                    Megamicros.Systems['mu1024']['product_id'],
                    Megamicros.Systems['mu1024']['bus_address'],
                    Megamicros.Systems['mu1024']['endpoint_in']
                )

            else:
                raise MuException("No MegaMicros device found")

        except MuException as e:
            log.error(f"Error during MegaMicros device check and opening: {e}")
            raise

    def wait( self, duration: float=0 ) -> None:  
        """ Wait for a given duration in seconds

        Parameters:
        -----------
        duration : float
            The waiting duration in seconds
        """
        if duration > 0:
            time.sleep( duration )

        self.usb.asyncBulkTransferWait()
        self.usb.release()
        log.info( " .Megamicros acquisition process successfully ended" )

    def stop( self ) -> None:
        """ Stop data acquisition on the MegaMicros device
        """
        log.info( " .Stopping Megamicros acquisition process" )
        self.usb.asyncBulkTransferStop()
        self.usb.release()

    def run( self, *args, **kwargs ) -> None:
        """ Run data acquisition on the MegaMicros device through USB async bulk transfer
        """

        if len( args ) > 0:
            raise MuException( f"Run() method does not accept direct arguments" )

        self._set_run_settings( [], kwargs=kwargs )

        # 1024 systems not yet implemented
        if not self.__system['name'] in ['mu32', 'mu32usb2', 'mu32a', 'mu256', 'mu256h']:
            raise MuException(f"MegaMicros system {self.__system['name']} not supported yet for data acquisition")

        if self.channels_number == 0 :
            raise MuException("No channel activated for data acquisition")

        # Set USB configuration
        self.usb.setBuffersNumber( self.__usb_buffers_number )
        self.usb.setBufferSize( self.frame_length * self.channels_number * TRANSFER_DATAWORDS_SIZE )

        # verbose
        log.info( f" .Starting run execution on Megamicros device..." )
        log.info( f"  > Run infinite loop (duration=0)" if self.duration == 0 else f"  > Perform {self.duration}s run loop" )
        log.info( f"  > Sampling frequency: {self.sampling_frequency} Hz" )
        log.info( f"  > FPGA clockdiv value: {self.clockdiv}" )
        log.info( f"  > {len(self.mems)} activated microphones" )
        log.info( f"  > Activated microphones: {self.mems}" )
        log.info( f"  > MEMs sensibility: {self.mems_sensibility}" )
        log.info( f"  > {len(self.analogs)} activated analogic channels" )
        log.info( f"  > Activated analogic channels: {self.analogs }" )
        log.info( f"  > Analogics sensibility: {self.analogs_sensibility}" )
        log.info( f"  > Whether counter is activated: {'YES' if 0 in self.counters else 'NO'}" )
        log.info( f"  > Whether status is activated: {'YES' if 0 in self.status else 'NO'}" )
        log.info( f"  > Time activation (MEMS powering delay): {self.time_activation} ms" )
        log.info( f"  > Total channels number is {self.channels_number}" )
        log.info( f"  > Datatype: {str( self.datatype )}" )
        log.info( f"  > Frame length in samples number: {self.frame_length} samples" )
        log.info( f"  > Frame duration: {self.frame_duration} s ({self.frame_duration * 1000} ms)" )
        log.info( f"  > Frame size in bytes: {self.frame_length * self.channels_number * TRANSFER_DATAWORDS_SIZE}" )
        log.info( f"  > Number of USB transfer buffers: {self.__usb_buffers_number}" )
        log.info( f"  > USB queue length: {'infinite queuing' if self.queue_length == 0 else self.queue_length} (frames)" )
        log.info( f"  > USB queue size: {'infinite queuing' if self.usb.queue_size == 0 else self.usb.queue_size} (bytes)" )
        log.info( f"  > Starting from external triggering: {'True' if self.__start_trigg_status else 'False'}" )
        log.info( f"  > Local H5 recording {'on' if self.h5_recording else 'off'}" )

        # Claim USB interface
        self.usb.claim()

        # set device configuration
        try:
            self.__ctrlResetMu()
            self.__ctrlClockdiv( self.clockdiv, self.time_activation / 1000 )
            self.__ctrlTixels( 0 )
            self.__ctrlDatatype( self.datatype )
            self.__ctrlMems( request='activate', mems=self.mems )
            self.__ctrlCSA( counter=True if 0 in self.counters else False, status=True if 0 in self.status else False, analogs=self.analogs )
            self.__ctrlStart()

        except Exception as e:
            log.error(f"Error during device configuration: {e}")
            self.usb.release()
            raise

        # Run data acquisition through USB async bulk transfer
        try:
            self.usb.asyncBulkTransfer(self.duration)
        except Exception as e:
            log.error(f"Error during USB async bulk transfer: {e}")
            self.usb.release()
            raise


    def setSamplingFrequency( self, sampling_frequency: int ) -> None:
        """ Set the sampling frequency for next runs on the device. The clockdiv property is updated accordingly.
        
        Parameters:
        -----------
        sampling_frequency : float
            The sampling frequency (default is 50kHz or 48 kHz on Aikhous systems)
        """
        super().setSamplingFrequency(sampling_frequency)
        self.__clockdiv = ( self.__clock_divider_reference // self.sampling_frequency ) - 1

    def setClockdiv( self, clockdiv: int ) -> None:
        """ Set the clockdiv that state for the sampling frequency for next runs on the device. 
        The sampling frequency property is updated accordingly.

        Parameters:
        -----------
        clockdiv : int
            The clock divider (default is 9 for 50kHz or 48 kHz on Aikhous systems)
        """
        self.__clockdiv = clockdiv
        self.setSamplingFrequency(self.__clock_divider_reference // ( self.__clockdiv + 1 ))

    def setClockDividerReference( self, clock_divider_reference: int ) -> None:
        """ Set the clock divider reference that state for the sampling frequency for next runs on the device.
        The sampling frequency property is updated accordingly.
        This can be used for Aikhous systems that use a 480kHz clock divider reference instead of 500kHz for usual systems.

        Parameters:
        -----------
        clock_divider_reference : int
            The clock divider reference (default is 500kHz for usual systems, 480kHz for Aikhous systems)
        """
        self.__clock_divider_reference = clock_divider_reference
        self.setSamplingFrequency(self.__clock_divider_reference // ( self.__clockdiv + 1 ))

    def setStartTriggStatus( self, trigger_status: bool ) -> None:
        """ Set the start trigger status

        Parameters
        ----------
        trigger_status: bool
            The start trigger status (True for external hard trigger, False for internal soft trigger)
        """
        self.__start_trigg_status = trigger_status

    def setUsbBuffersNumber( self, number: int ) -> None:
        """ Set the USB buffers number used in USB the data acquisition process.
        Should never be set to less than 4 to ensure proper data transfer without overflow. 8 is a good default value.

        Parameters
        ----------
        number: int
            The USB buffers number
        """
        self.__usb_buffers_number = number

    def setCounter( self, counters: list[int] | bool ) -> None:
        """ Activate counters channels. Overload the parent setCounter method to considere multiple counters.
        """

        if isinstance( counters, bool ):
            if counters:
                self.__counters = [0]
            else:
                self.__counters = []
        else:
            self.__counters = counters

    def setCounters( self, counters: list[int] | bool ) -> None :
        """ Activate counters channels
        """

        if isinstance( counters, bool ):
            if counters:
                self.__counters = [0]
            else:
                self.__counters = []
        else:
            self.__counters = counters

    def setStatus( self, status: list[int] | bool) -> None :
        """ Activate status channels
        """

        if isinstance( status, bool ):
            if status:
                self.__status = [0]
            else:
                self.__status = []
        else:
            self.__status = status

    def setAnalogsSensibility( self, sensibility: float ) -> None:
        """ Set analogs sensibility
        
        Parameters:
        -----------
        sensibility : float
            The analogs sensibility in V/digit
        """

        self.__analogs_sensibility = sensibility

    def setTimeActivation( self, time_activation: int ) -> None:
        """ Set time activation after MEMs powering in milliseconds
        
        Parameters:
        -----------
        time_activation : int
            time activation after MEMs powering in milliseconds
        """

        self.__time_activation = time_activation

    def _set_run_settings( self, args, kwargs ) -> None :
        """ Set settings for run method
        
        Parameters
        ----------
        args: array
            direct arguments of the run function
        kwargs: array
            named arguments of the run function
        """
        
        if len( args ) > 0:
            log.warning( f" .Direct arguments are not accepted. Use named arguments instead ({args})" )
            raise MuException( "Direct arguments are not accepted" )

        try:
            if 'analogs' in kwargs:
                self.setActiveAnalogs( kwargs['analogs'] )
                del kwargs['analogs']

            if 'available_analogs' in kwargs:
                self.setAvailableAnalogs( kwargs['available_analogs'] )
                del kwargs['available_analogs']

            if 'counter' in kwargs:
                if isinstance( kwargs['counter'], list ) or isinstance( kwargs['counter'], tuple ):
                    self.setCounters( kwargs['counter'] )
                else:   
                    self.setCounter( kwargs['counter'] )
                del kwargs['counter']

            if 'counters' in kwargs:
                if isinstance( kwargs['counters'], list ) or isinstance( kwargs['counters'], tuple ):
                    self.setCounters( kwargs['counters'] )
                else:   
                    self.setCounter( kwargs['counters'] )
                del kwargs['counters']

            if 'sampling_frequency' in kwargs:
                self.setSamplingFrequency( kwargs['sampling_frequency'] )
                del kwargs['sampling_frequency']

            if 'clockdiv' in kwargs:
                self.setClockdiv( kwargs['clockdiv'] )
                del kwargs['clockdiv']

            if 'mems_sensibility' in kwargs:
                self.setMemsSensibility( kwargs['mems_sensibility'] )
                del kwargs['mems_sensibility']  

            if 'analogs_sensibility' in kwargs:
                self.setAnalogsSensibility( kwargs['analogs_sensibility'] )

            if 'queue_length' in kwargs:
                self.setQueueLength( kwargs['queue_length'] )
                del kwargs['queue_length']

            if 'time_activation' in kwargs:
                self.setTimeActivation( kwargs['time_activation'] )
                del kwargs['time_activation']

            super()._set_run_settings( [], kwargs )

        except Exception as e:
            raise MuException( f"Run failed on settings: {e}")


    def selftest(self, duration: int = DEFAULT_SELFTEST_DURATION) -> tuple[list[float], list[float]]:
        """
        @brief Perform a one-second recording test to obtain and update the Megamicros receiver settings.
        The test is performed with the default settings, not with the current settings. 
        Current settings are then updated with the tests results. 
        Check mems, analogs, status and counter channels.
        Update _available_mems and _available_analogs and others.
        The system should be detected before. 

        @param duration: duration of the selftest in seconds
        """

        # Compute MEMs energy using the bulk transfer method because the bulkRead method does not work on Windows
        if platform.system() == 'Windows':
            log.info(" .Performing selftest using bulk transfer method...")
            raise NotImplementedError( 'Usb.asyncBulkTransferWait() not implemented yet' )

        if not self.usb.isOpened():
            raise MuException("MegaMicros device is not opened")

        log.info(" .Performing selftest using bulk read method...")

        self.usb.claim()
        try:
            mems_number = self.__system['beams'] * MU_BEAM_MEMS_NUMBER
            analogs_number = self.__system['analogs']
            channels_number = mems_number + analogs_number + self.__system['counters'] + self.__system['status']
            frame_length = int( duration * self.sampling_frequency )
            buffer_size = channels_number * frame_length * TRANSFER_DATAWORDS_SIZE
            self.__ctrlResetMu()
            self.__ctrlClockdiv( DEFAULT_CLOCKDIV, DEFAULT_TIME_ACTIVATION / 1000 )
            self.__ctrlTixels( frame_length )
            self.__ctrlDatatype( 'int32' )
            self.__ctrlMems( request='activate', mems='all' )
            self.__ctrlCSA( counter=True, status=True, analogs='all' )
            self.__ctrlStart()
            bytes_read:bytearray = self.usb.syncBulkRead( 
                size=buffer_size, 
                time_out=USB_DEFAULT_TRANSFER_TIMEOUT 
            )
            self.__ctrlStop()

            if len( bytes_read ) != buffer_size:
                self.usb.release()
                raise MuException( f'Selftest failed: expected {buffer_size} bytes but got {len( bytes_read )} bytes' )

            log.info(" .Selftest completed successfully")

        except MuException as e:
            log.error(f"Error during selftest: {e}")
            self.usb.release()
            raise

        self.usb.release()

        # Check data length
        data = np.frombuffer( bytes_read, dtype=np.int32 )
        if len( data ) != frame_length * channels_number:
            # Windows platform failed on this test (Zadig driver issue ?)
            # Note that Zadig driver is based on libusb 0.1 porting for Windows while Python libusb is based on libusb 1.0
            if platform.system() == 'Windows':
                log.warning( f"Received {len(data)} data bytes instead of {frame_length * channels_number} ({frame_length} samples)" )
                if len( data ) > frame_length * channels_number:
                    log.warning( f"Windows platform detected --> Data length will be adjusted to {frame_length} samples." )
                    data = data[:frame_length * channels_number]
                else:
                    new_data_length = len( data ) // channels_number
                    log.warning( f"Windows platform detected --> Data length will be adjusted to {new_data_length*channels_number} ({new_data_length} samples)." )
                    if len( data ) % channels_number == 0:
                        log.warning( f"Removed exactly {frame_length - new_data_length} samples of {channels_number} channels each." )
                    else:
                        log.warning( f"Removed {frame_length * channels_number - len( data )} bytes than cannot be expressed as multiples of channels or samples number." )
                    data = data[:new_data_length * channels_number]
                    data_length = new_data_length
            else:
                raise MuException( f"Received {len(data)} data bytes instead of {frame_length * channels_number}" )
        

        # convert to float with sensibility factor
        if self.__system['name'] in ['mu32', 'mu32usb2', 'mu32a', 'mu256', 'mu256h']:

            # Reshape data to (channels, samples)
            data = data.reshape( ( channels_number, frame_length ), order='F' )

            # Get signals with sensibility factor
            mems_signal = data[1:mems_number+1].astype( np.float32 ) * self.mems_sensibility
            analogs_signal = data[mems_number+1:mems_number+analogs_number+1].astype( np.float32 ) * self.analogs_sensibility

            # Compute mean energy  
            mems_power = np.sum( mems_signal**2, axis=1 ) / frame_length
            if analogs_number > 0:
                analogs_power = np.sum( analogs_signal**2, axis=1 ) / frame_length
            else:
                analogs_power = np.array([])
                
            mems_power = np.where( mems_power > 0, 1, 0 )
            analogs_power = np.where( analogs_power > 1e-10 , 1, 0 )

            log.info( f" .Autotest results:" )
            log.info( f"  > equivalent recording time is: {frame_length / self.sampling_frequency} seconds" )
            log.info( f"  > Received {len(bytes_read)} data bytes: {frame_length} samples on {channels_number} channels")
            log.info( f"  > detected {len( np.where( mems_power > 0 )[0] )} active MEMs: {np.where( mems_power > 0 )[0].tolist()}" )
            if analogs_number > 0:
                log.info( f"  > detected {len( np.where( analogs_power > 0 )[0] )} active analogs: {np.where( analogs_power > 0 )[0].tolist()}" )
            else:
                log.info( f"  > detected no active analogs" )
            log.info( f"  > detected counter channel with values from {data[0][0]} to {data[0][-1]}" )
            log.info( f"  > estimated data lost: {data[0][-1] - data[0][0] + 1 - frame_length} samples" )
            log.info( f"  > detected status channel with values {data[channels_number-1][0]} <-> {data[channels_number-1][-1]}" )
            log.info( f" .Selftest endded successfully" )

            # Set available mems and analogs according to selftest results
            self.setAvailableMems(np.where(np.array(mems_power) > 0)[0].tolist())
            self.setAvailableAnalogs(np.where(np.array(analogs_power) > 0)[0].tolist())

            return mems_power.tolist(), analogs_power.tolist()
        else:
            raise MuException( f"Selftest not implemented for system type {self.__system['name']}" )
            


    def __ctrlWrite(self, request: int, data: bytes=b"", timeout=USB_DEFAULT_WRITE_TIMEOUT) -> None:
        """
        Send a control write USB request to the MegaMicros device

        Parameters
        ----------
        request: str
            USB request code
        data: bytes, optional
            USB request data (default to empty)

        Raises
        ------
        MuException
            In case of error during the USB transfer
        """
        try:
            if data == b"":
                self.__usb.ctrlWriteReset(request, timeout)
            else:
                self.__usb.ctrlWrite(request, data, timeout)
        
        except MuException as e:
            log.error(f"Error during control write request {request:#04x}: {e}")
            raise


    def __ctrlWriteReset(self, request: int, timeout=USB_DEFAULT_WRITE_TIMEOUT) -> None:
        """
        Send a control write USB request to the MegaMicros device with no data

        Parameters
        ----------
        request: str
            USB request code

        Raises
        ------
        MuException
            In case of error during the USB transfer
        """

        try:
            self.__usb.ctrlWriteReset(request, timeout)
        except MuException as e:
            log.error(f"Error during control write reset request {request:#04x}: {e}")
            raise


    def __ctrlTixels( self, samples_number ):
        """
        Set the samples number to be sent by the Megamicros system 
        """

        buf = create_string_buffer( 5 )
        buf[0] = MU_CMD_COUNT
        buf[1] = bytes(( samples_number & 0x000000ff, ) )
        buf[2] = bytes( ( ( ( samples_number & 0x0000ff00 ) >> 8 ),) )
        buf[3] = bytes( ( ( ( samples_number & 0x00ff0000 ) >> 16 ),) )
        buf[4] = bytes( ( ( ( samples_number & 0xff000000 ) >> 24 ),) )
        self.__ctrlWrite( 0xB4, buf, USB_DEFAULT_WRITE_TIMEOUT )


    def __ctrlResetAcq( self ):
        """
        Reset and purge fifo
        No documention found about the 0x06 code value. Use ctrlResetMu() instead for a complete system reset
        """
        buf = create_string_buffer( 1 )
        buf[0] = MU_CMD_RESET
        self.__ctrlWrite( 0xB0, buf, USB_DEFAULT_WRITE_TIMEOUT )
        buf[0] = MU_CMD_PURGE
        self.__ctrlWrite( 0xB0, buf, USB_DEFAULT_WRITE_TIMEOUT )


    def __ctrlResetFx3( self ):
        """
        Mu32 needs the 0xC4 command but not the 0xC2 unlike what is used on other programs...
        Regarding the Mu32 documentation, this control seems incomplete (/C0/C4/(B0 00)). 
        256 doc says that ctrlResetMu() is the complete sequence that should be used with fiber (/C0/C4/(B0 00)/C4/C0)
        while ctrlResetFx3() should only be used with USB with non-fiber USB.
        Please use ctrlResetMu() in all cases
        """
        try:
            self.__ctrlWriteReset( MU_CMD_FX3_RESET, USB_DEFAULT_WRITE_TIMEOUT )
            self.__ctrlWriteReset( MU_CMD_FX3_PH, USB_DEFAULT_WRITE_TIMEOUT )
        except Exception as e:
            log.error( f"Fx3 reset failed: {e}" ) 
            raise


    def __ctrlResetMu( self ):
        """
        full reset of Mu32 receiver using fiber or not
        """
        buf = create_string_buffer( 1 )
        buf[0] = MU_CMD_RESET
        try:
            self.__ctrlWriteReset( MU_CMD_FX3_RESET, USB_DEFAULT_WRITE_TIMEOUT )
            time.sleep( DEFAULT_TIME_ACTIVATION_RESET / 1000)
            self.__ctrlWriteReset( MU_CMD_FX3_PH, USB_DEFAULT_WRITE_TIMEOUT )
            time.sleep( DEFAULT_TIME_ACTIVATION_RESET / 1000)
            self.__ctrlWrite( MU_CMD_FPGA_0, buf, USB_DEFAULT_WRITE_TIMEOUT )
            time.sleep( DEFAULT_TIME_ACTIVATION_RESET / 1000)
            self.__ctrlWriteReset( MU_CMD_FX3_PH, USB_DEFAULT_WRITE_TIMEOUT )
            time.sleep( DEFAULT_TIME_ACTIVATION_RESET / 1000)
            self.__ctrlWriteReset( MU_CMD_FX3_RESET, USB_DEFAULT_WRITE_TIMEOUT )
            time.sleep( DEFAULT_TIME_ACTIVATION_RESET / 1000)
        except Exception as e:
            log.error( f"Mu32 reset failed: {e}" ) 
            raise

    def __ctrlResetFPGA( self ):
        """
        reset of FPGA
        """
        buf = create_string_buffer( 1 )
        buf[0] = MU_CMD_RESET
        try:
            self.__ctrlWrite( MU_CMD_FPGA_0, buf, USB_DEFAULT_WRITE_TIMEOUT )
        except Exception as e:
            log.error( f"FPGA reset failed: {e}" ) 
            raise


    def __ctrlClockdiv( self, clockdiv=0x09, time_activation=DEFAULT_TIME_ACTIVATION / 1000 ):
        """
        Init acq32: set sampling frequency and supplies power to microphones 
        """
        buf = create_string_buffer( 2 )
        buf[0] = MU_CMD_INIT
        buf[1] = clockdiv
        try:
            self.__ctrlWrite( MU_CMD_FPGA_1, buf, USB_DEFAULT_WRITE_TIMEOUT )
        except Exception as e:
            log.error( f"Mu32 clock setting and powerwing on microphones failed: {e}" ) 
            raise	

        """
        wait for mems activation
        """
        time.sleep( time_activation )


    def __ctrlDatatype( self, datatype='int32' ):
        """
        Set datatype
        ! note that float32 is not considered -> TO DO
        """ 
        buf = create_string_buffer( 2 )
        buf[0] = MU_CMD_DATATYPE
        if datatype=='int32':
            buf[1] = MU_CODE_DATATYPE_INT32
        elif datatype=='float32':
            buf[1] = MU_CODE_DATATYPE_FLOAT32
        else:
            raise MuException( 'Mu32::ctrlDatatype(): Unknown data type [%s]. Please, use [int32] or [float32]' % datatype )

        try:
            self.__ctrlWrite( MU_CMD_FPGA_1, buf, USB_DEFAULT_WRITE_TIMEOUT )
        except Exception as e:
            log.error( f"Mu32 datatype setting failed: {e}" ) 
            raise	


    def __ctrlMems( self, request:str, mems:str|list|tuple ='all' ):
        """
        Activate or deactivate MEMs

        Parameters
        ----------
        request: str
            The request type: activate or deactivate
        mems: str or array, optional
            The MEMs to activate or deactivate (default is 'all')
        """

        try:
            buf = create_string_buffer( 4 )
            buf[0] = MU_CMD_ACTIVE		
            buf[1] = 0x00					# module
            pluggable_mems_beams = self.__system['beams']
            if mems == 'all':
                if request == 'activate':
                    for beam in range( pluggable_mems_beams ):
                        buf[2] = beam		# beam number
                        buf[3] = 0xFF		# active MEMs map
                        self.__ctrlWrite( MU_CMD_FPGA_3, buf, USB_DEFAULT_WRITE_TIMEOUT )
                elif request == 'deactivate':
                    for beam in range( pluggable_mems_beams ):
                        buf[2] = beam		
                        buf[3] = 0x00		
                        self.__ctrlWrite( MU_CMD_FPGA_3, buf, USB_DEFAULT_WRITE_TIMEOUT )
                else:
                    raise MuException( 'Megamicros::ctrlMems(): Unknown parameter [%s]' % request )
            else:
                if request == 'activate':
                    map_mems = [0 for _ in range( pluggable_mems_beams )]
                    for mic in mems:
                        mic_index = mic % MU_BEAM_MEMS_NUMBER
                        beam_index = int( mic / MU_BEAM_MEMS_NUMBER )
                        if beam_index >= pluggable_mems_beams:
                            raise MuException( 'microphone index [%d] is out of range (should be less than %d)' % ( mic,  pluggable_mems_beams * MU_BEAM_MEMS_NUMBER ) )
                        map_mems[beam_index] += ( 0x01 << mic_index )

                    for beam in range( pluggable_mems_beams ):
                        if map_mems[beam] != 0:
                            buf[2] = beam
                            buf[3] = map_mems[beam]				
                            self.__ctrlWrite( MU_CMD_FPGA_3, buf, USB_DEFAULT_WRITE_TIMEOUT )
                else:
                    raise MuException( 'Megamicros::ctrlMems(): request [%s] is not implemented' % request )
        except Exception as e:
            log.error( f"Megamicros microphones activating failed: {e}" ) 
            raise	


    def __ctrlCSA( self, counter: bool, status: bool, analogs: str|list|tuple='all' ):
        """
        Activate or deactivate analogic, status and counter channels

        Parameters
        ----------
        counter: bool
            Activate or deactivate counter channel
        status: bool
            Activate or deactivate status channel
        analogs: list or tuple
            Activate or deactivate analogic channels
        """		

        pluggable_analogs_number = self.__system['analogs']
        if analogs == 'all':
            analogs = [i for i in range( pluggable_analogs_number)]
            
        buf = create_string_buffer( 4 )
        buf[0] = MU_CMD_ACTIVE		# command
        buf[1] = 0x00				# module
        buf[2] = 0xFF				# counter, status and analogic channels

        map_csa = 0x00
        if len( analogs ) > 0:
            for anl_index in analogs:
                map_csa += ( 0x01 << anl_index ) 
        if status:
            map_csa += ( 0x01 << 6 )
        if counter:
            map_csa += ( 0x01 << 7 )

        buf[3] = map_csa

        try:
            self.__ctrlWrite( MU_CMD_FPGA_3, buf, USB_DEFAULT_WRITE_TIMEOUT )
        except Exception as e:
            log.error( f"Mu32 analogic channels and status activating failed: {e}" ) 
            raise	


    def __ctrlStart( self ):
        """
        start acquiring by soft triggering
        """
        buf = create_string_buffer( 2 )
        buf[0] = MU_CMD_START
        buf[1] = 0x00

        try:
            self.__ctrlWrite( MU_CMD_FPGA_1, buf, USB_DEFAULT_WRITE_TIMEOUT )
        except Exception as e:
            log.error( f"Mu32 starting failed: {e}" ) 
            raise	

    def __ctrlStartTrig( self ):
        """
        start acquiring by external triggering
        """
        buf = create_string_buffer( 2 )
        buf[0] = MU_CMD_START
        buf[1] = 0x01										# front montant 
        #buf[1] = 0x01 + ( 0x01 << 7 )						# (état haut)

        try:
            self.__ctrlWrite( MU_CMD_FPGA_1, buf, USB_DEFAULT_WRITE_TIMEOUT )
        except Exception as e:
            log.error( f"Mu32 starting by external trig failed: {e}" ) 
            raise	

    def __ctrlStop( self ):
        """
        stop acquiring by soft triggering
        """
        buf = create_string_buffer( 2 )
        buf[0] = MU_CMD_STOP
        buf[1] = 0x00

        try:
            self.__ctrlWrite( MU_CMD_FPGA_1, buf, USB_DEFAULT_WRITE_TIMEOUT )
        except Exception as e:
            log.error( f"Mu32 stop failed: {e}" ) 
            raise


    def __ctrlPowerOffMic( self ):
        """
        powers off microphones
        """
        buf = create_string_buffer( 2 )
        buf[0] = MU_CMD_RESET

        try:
            self.__ctrlWrite( MU_CMD_FPGA_0, buf, USB_DEFAULT_WRITE_TIMEOUT )
        except Exception as e:
            log.error( f"Mu32 microphones powering off failed: {e}" ) 
            raise	

    def __iter__( self ) :
        """ Init iterations over the antenna data """

        self.__it = 0
        return self

    def __next__( self ) -> np.ndarray|StopIteration :
        """ next iteration over the antenna data 
        """
        
        try:
            data = self.usb.queue.get( timeout=USB_DEFAULT_QUEUE_TIMEOUT )
            # format data according to datatype
            if self.datatype == 'int32':
                data = np.frombuffer( data, dtype=np.int32 )
                data = data.reshape( ( -1, self.frame_length ), order='F' )
            elif self.datatype == 'float32':
                data = np.frombuffer( data, dtype=np.float32 )
                data = data.reshape( ( -1, self.frame_length ), order='F' )
            else:
                raise MuException( f"Unknown datatype {self.datatype}" )
            self.__it += 1
            return data

        except queue.Empty:
            raise StopIteration

    def __len__( self) -> int : 
        """ Get the number of elemnts in the queue """

        if self.usb.queue is None:
            return 0
        else:
            return self.usb.queue_content



# megamicros.acoustics.omp.py
#
# ® Copyright 2024-2025 Bimea
# Author: francois.ollivier@bimea.io
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Orthogonal Matching Pursuit (OMP) for sparse signal approximation.

Features:
    - Greedy algorithm for selecting dictionary atoms
    - Supports non-negative coefficients
    - Configurable stopping criteria: max coefficients, max iterations, tolerance
    - Returns detailed result object with solution and runtime parameters

Notes:
    
Orthogonal Matching Pursuit (OMP) is a greedy algorithm used for sparse signal approximation. 
It's often used in the field of signal processing and machine learning for tasks such as feature selection and sparse coding.

The algorithm works by iteratively selecting the dictionary elements (also known as "atoms") 
that are most correlated with the current residual (the difference between the original signal and the current approximation), 
and then updating the approximation and residual.

Here's a simplified description of the algorithm:
* Initialize the residual with the original signal and the approximation with zero.
* Find the dictionary atom that is most correlated with the current residual.
* Update the approximation by projecting the original signal onto the space spanned by the selected atoms.
* Update the residual by subtracting the updated approximation from the original signal.
* Repeat steps 2-4 until a stopping criterion is met (e.g., a maximum number of atoms have been selected, or the residual is below a certain threshold).

The "orthogonal" in OMP refers to the fact that the residual is kept orthogonal to the space spanned by the selected atoms, 
which leads to a more stable and accurate approximation compared to some other greedy algorithms.

The OMP algorithm has several advantages over other feature selection methods:

* Simplicity and efficiency: OMP is a “greedy” algorithm that selects, one by one, the features most correlated with the current residual.
This makes it relatively simple to understand and implement, and efficient in terms of computation time.
* Stability: OMP keeps the residual orthogonal to the space generated by the selected features, 
which leads to a more stable and accurate approximation compared to some other greedy methods.
* Support for sparsity: OMP is particularly well suited to problems where the solution is sparse 
(i.e., it contains many zeros). 
It is capable of recovering the exact solution in the presence of noise, provided that the signal is sufficiently sparse.
* Non-negativity: OMP can be easily adapted to impose a non-negativity constraint on the coefficients, 
which can be useful in certain contexts (e.g., 
when the features represent physical quantities that cannot be negative).

However, like any algorithm, OMP also has its limitations. 
For example, it may not perform as well when the signal is not sparse, 
or when the features are highly correlated with each other.
In these cases, it may select one feature at the expense of another that is nearly identical.
* No regularization: Unlike other methods like Lasso or Ridge, OMP does not have a regularization mechanism to control model complexity. 
This can lead to overfitting if the number of features is very large compared to the number of samples.
* Choice of number of features: OMP requires specifying the number of features to select, 
which can be difficult to determine in advance. 
Some variants of OMP allow for automatic determination of this number, but they may be more complex to implement.


Examples:
    Basic usage::

        Comming soon...

    Advanced usage::

        See the Notebooks for advanced usage examples.

Documentation:
    Full MegaMicros documentation is available at: https://readthedoc.bimea.io
"""

import numpy as np
from scipy.optimize import nnls
from ..log import log
from ..exception import MuException

class Result(object):
    """Result object for storing input and output data for omp.  When called from 
    `omp`, runtime parameters are passed as keyword arguments and stored in the 
    `params` dictionary.

    Attributes:
        X:  Predictor array after (optional) standardization.
        y:  Response array after (optional) standarization.
        ypred:  Predicted response.
        residual:  Residual vector.
        coef:  Solution coefficients.
        active:  Indices of the active (non-zero) coefficient set.
        err:  Relative error per iteration.
        params:  Dictionary of runtime parameters passed as keyword args.   
    """
    
    def __init__(self, **kwargs):
        
        # to be computed
        self.X = None
        self.y = None
        self.ypred = None
        self.residual = None
        self.coef = None
        self.active = None
        self.err = None
        
        # runtime parameters
        self.params = {}
        for key, val in kwargs.items():
            self.params[key] = val
            
    def update(self, coef, active, err, residual, ypred):
        """Update the solution attributes.
        """
        self.coef = coef
        self.active = active
        self.err = err
        self.residual = residual
        self.ypred = ypred


def omp( X, y, nonneg=True, ncoef=None, maxit=10, tol=1e-3, ztol=1e-12, verbose=False ):
    """Compute sparse orthogonal matching pursuit solution with unconstrained
    or non-negative coefficients.
    
    Parameters
    ----------
    X: dict
        Dictionary array of size n_samples x n_features. 
    y: np.array 
        Reponse array of size n_samples x 1.
    nonneg: boolean
        Enforce non-negative coefficients. Default is True.
    ncoef: int
        Max number of coefficients.  Set to n_features/2 by default.
    tol: float
        Convergence tolerance.  If relative error is less than tol * ||y||_2, exit.
    ztol: float
        Residual covariance threshold.  If all coefficients are less than ztol * ||y||_2, exit.
    verbose: Boolean
        print some info at each iteration.
        
    Returns
    -------
    result:  Result object.  
        See Result.__doc__ for more info.
    """
    
    def norm2(x):
        return np.linalg.norm(x) / np.sqrt(len(x))
    
    # initialize result object
    result = Result(nnoneg=nonneg, ncoef=ncoef, maxit=maxit, tol=tol, ztol=ztol)
    if verbose:
        print(result.params)
    
    # check types, try to make somewhat user friendly
    if type(X) is not np.ndarray:
        X = np.array(X)
    if type(y) is not np.ndarray:
        y = np.array(y)
        
    # check that n_samples match
    if X.shape[0] != len(y):
        raise MuException( f"OMP: X and y must have same number of rows (samples)" )
    
    # store arrays in result object    
    result.y = y
    result.X = X
    
    # for rest of call, want y to have ndim=1
    if np.ndim(y) > 1:
        y = np.reshape(y, (len(y),))
        
    # by default set max number of coef to half of total possible
    if ncoef is None:
        ncoef = int(X.shape[1]/2)
    
    # initialize things
    X_transpose = X.T                        # store for repeated use
    #active = np.array([], dtype=int)         # initialize list of active set
    active = []
    coef = np.zeros(X.shape[1], dtype=float) # solution vector
    residual = y                             # residual vector
    ypred = np.zeros(y.shape, dtype=float)
    ynorm = norm2(y)                         # store for computing relative err
    err = np.zeros(maxit, dtype=float)       # relative err vector
    
    # Check if response has zero norm, because then we're done. This can happen
    # in the corner case where the response is constant and you normalize it.
    if ynorm < tol:     # the same as ||residual|| < tol * ||residual||
        log.warning( 'OMP: Norm of the response is less than convergence tolerance.' )
        result.update(coef, active, err[0], residual, ypred)
        return result
    
    # convert tolerances to relative
    tol = tol * ynorm       # convergence tolerance
    ztol = ztol * ynorm     # threshold for residual covariance
    
    log.info(' .OMP: Iteration, relative error, number of non-zeros')
   
    # main iteration
    for it in range(maxit):
        
        # compute residual covariance vector and check threshold
        rcov = np.dot(X_transpose, residual)
        if nonneg:
            i = np.argmax(rcov)
            rc = rcov[i]
        else:
            i = np.argmax(np.abs(rcov))
            rc = np.abs(rcov[i])
        if rc < ztol:
            log.warning('OMP: All residual covariances are below threshold.')
            break
        
        # update active set
        if i not in active:
            #active = np.concatenate([active, [i]], axis=1)
            active.append(i)
            
        # solve for new coefficients on active set
        if nonneg:
            #coefi, _ = nnls(X[:, active], y)
            # ComplexWarning: Casting complex values to real discards the imaginary part -> explicit convertion to float64:
            coefi, _ = nnls(X[:, active].real, y.real)
        else:
            coefi, _, _, _ = np.linalg.lstsq(X[:, active], y)

        coef[active] = coefi   # update solution
        
        # update residual vector and error
        residual = y - np.dot(X[:,active], coefi)
        ypred = y - residual
        err[it] = norm2(residual) / ynorm  
        
        # print status
        log.info( f" > {it}, {err[it]}, {len(active)}" )
            
        # check stopping criteria
        if err[it] < tol:  # converged
            log.info(' .OMP: Converged.')
            break
        if len(active) >= ncoef:   # hit max coefficients
            log.info(' .OMP: Found solution with max number of coefficients.')
            break
        if it == maxit-1:  # max iterations
            log.info(' .OMP: Hit max iterations.')
    
    result.update(coef, active, err[:(it+1)], residual, ypred)
    return result

if __name__ == '__main__':
    pass



# megamicros.exception.py
#
# ® Copyright 2024-2025 Bimea
# Author: bruno.gas@bimea.io
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Megamicros exception base class.

Features:
    - Base exception class for Megamicros library
    
Documentation:
    Full MegaMicros documentation is available at: https://readthedoc.bimea.io
"""


from megamicros.log import log 

class MuException( Exception ):
    """Exception base class for Megamicros """

    def __init__( self, message: str="" ):
        super().__init__( message )



# megamicros.apps.megamicros.py is the command line interface for megamicros.
#
# ® Copyright 2024-2025 Bimea
# Author: bruno.gas@bimea.io
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import argparse

from megamicros import __version__, welcome_msg
from megamicros.log import log
from megamicros.exception import MuException
from megamicros.core.mu import Megamicros

def arg_parse() -> tuple:

    parser = argparse.ArgumentParser()
    parser.add_argument( "-v", "--version", help=f"show megamicros installed version", action='store_true' )
    parser.add_argument( "--verbose", help=f"set verbose mode on", action='store_true' )
    parser.add_argument( "--check-usb", help=f"check usb device", action='store_true' )
    parser.add_argument( "--check-device", help=f"check megamicros device", action='store_true' )

    return parser.parse_args()

def main():

    args = arg_parse()

    if args.version:
        print( f"megamicros {__version__}" )
        return
    
    if args.verbose:
        log.setLevel( "INFO" )
    else:
        log.setLevel( "ERROR" )

    # Print welcome message
    print( welcome_msg )
    print( f"megamicros {__version__}" )

    if args.check_usb:
        print( "Checking USB..." )
        try:
            print( "Not implemented yet" )
            return
            #
            #devices = Megamicros.check_device()
            #print( f"Found following device {devices['usb_vendor_id']:04x}:{devices['usb_vendor_product']:04x} characteristics:")
            #print( f"  - system_type: {devices['system_type']}" )
            #print( f"  - usb_vendor_id: {devices['usb_vendor_id']:04x}" )
            #print( f"  - usb_vendor_product: {devices['usb_vendor_product']:04x}" )
            #print( f"  - usb_bus_address: {devices['usb_bus_address']}" )
            #print( f"  - pluggable_mems_number: {devices['pluggable_beams_number']*8}" )
            #print( f"  - pluggable_analogs_number: {devices['pluggable_analogs_number']}" )

        except MuException as e:
            print( f"Failed: {e}")

    elif args.check_device:
        print( "Checking Megamicros device..." )
        try:
            antenna = Megamicros()
            available_mems, available_analogs = antenna.selftest(1)
            print( f"Found Megamicros device:")
            print( f"  - {len(available_mems)} available MEMs found")
            print( f"  - {len(available_analogs)} available analogs found")
            print( f"  - MEMs: {available_mems}" )
            print( f"  - Analogs: {available_analogs}" )

        except MuException as e:
            print( f"Failed: {e}")


if __name__ == "__main__":
	main()



# -*- coding: utf-8 -*-

"""
Module contenant toutes les fonctions pour
la calibration géométrique d'antenne

Created on Wed May 14 12:20:11 2014
@author: charles

Voir aussi: GéoCalib dans bimea/Manips
"""

from __future__ import division
import numpy as np
import scipy.signal as sig

import scipy.optimize as opt
import scipy as sp
import pylab as pl
from scipy.spatial.distance import squareform, pdist
from scipy import linalg as linalg

#import rpy2.robjects as ro
#from rpy2.robjects.numpy2ri import numpy2ri
#ro.conversion.py2ri = numpy2ri

#from rpy2.robjects.packages import importr

###############################################################################
#Util: Création d'une matrice de pondération pour MDS locale:
def makeWlocal(D, dmax):
    """
    Contruit une matrice de pondération pour MDS local.
    """
    W = np.zeros_like(D)
    W[np.where(D<=dmax)] = 1
    W[np.where(D>dmax)] = 0
    np.fill_diagonal(W, 0)
    return W

#MDS classique:
def cmds(D, Ndim):
    """
    Classical MDS: X = cmds(D, Ndim)

    Entrées
    -------
    D: (Nr,Nr) array numpy
        Matrice des distances mic-mic complète (ie avec symétrie)
    Ndim: int
        Dimension spatiale de l'antenne à calibrer
    
    Sortie
    ------
    X: (Nr,Ndim) array numpy 
        matrice des coordonnées de l'antenne
    """
    Nr = D.shape[0]
    #idxtril = np.tril_indices(Nr,k=0)
    
    J = np.identity(Nr) - np.ones((Nr,Nr)) / Nr
    B = -0.5 * np.dot(J, np.dot(D**2,J))
    eigenValues, eigenVectors = np.linalg.eig(B)
    idx = eigenValues.argsort()[::-1]
    eigenValues = eigenValues[idx]
    eigenVectors = eigenVectors[:,idx]
    
    return np.dot(eigenVectors[:,0:Ndim], np.diag(eigenValues[0:Ndim]**0.5))

#Smacof:
#def smacof(D, Ndim, typeMds = 'ratio', W = 'null' , init = 'null',
#           ties = 'primary', verbose = False, relax = False, modulus = 1, 
#           itmax = 1000, eps = 1e-6, SplineDegree = 2, SplineIntKnots = 2):
#    """
#    Smacof MDS: X = smacof(D, Ndim, typeMds, W, init, ties, verbose, ...)
#    
#    smacofSym function wrap from the R package of Smacof method.
#
#    Inputs
#    ------
#    D: (Nr,Nr) array numpy
#        Matrice des distances mic-mic complète (ie avec symétrie)
#    Ndim: int
#        Dimension de l'espace de sortie
#    typeMds: str
#        "ratio", "interval", "ordinal" ou "mspline"
#    W: 'null' ou (Nr,Nr) array numpy
#        Matrice de pondération de D
#    init: 'null' ou matrice d'initialisation
#        
#    ties: "primary", "secondary" ou "tertiary"
#        Utile pour le typeMds = "ordinal"
#    verbose: Boolean
#        
#    relax: Boolean
#        Relaxation lors de la majorisation
#    modulus: int
#        Number of smacof iterations per monotone regression call
#    itmax: int
#        Nb max d'itérations
#    eps: float
#        Convergence criterion
#    SplineDegree: int
#        Degree of the spline for "mspline" MDS type
#    SplineIntKnots: int
#        Number of interior knots of the spline for "mspline" MDS type
#        
#    
#    Outputs
#    -------
#    X: (Nr,Ndim) array numpy 
#        matrice des coordonnées de l'antenne
#    
#    Optimization Parameters: list (11 elements)
#        0 delta = diss
#        1 obsdiss = dhat
#        2 confdiss = confdiss
#        3 iord = dhat2$iord.prim,
#        4 stress = stress,
#        5 spp = spp,
#        6 ndim = p,
#        7 model = "Symmetric SMACOF",
#        8 niter = itel,
#        9 nobj = n,
#        10 type = type,
#        11 call = match.call()) 
#        
#    """
#    
#    rsmacof = importr('smacof')
#    
#    if W == 'null': W = ro.NULL
#    if init == 'null': init = ro.NULL
#     
#    stress = rsmacof.smacofSym(D, Ndim, typeMds, W, init, ties, verbose)
#    
#    output = np.array(stress[4]) # X Nr*Ndim
#    OptParams = [np.array(stress[0]), np.array(stress[1]), np.array(stress[2]),
#                 np.array(stress[3]), stress[5][0], np.array(stress[6]), 
#                stress[7][0], stress[8][0], stress[9][0], stress[10][0],
#                stress[11][0], stress[12]]
#    
#    return output, OptParams

#Robust MDS:
def smart_logspace(start, stop, num):
    startexp, stopexp = np.log10(start), np.log10(stop)
    return np.logspace(startexp, stopexp, num)

def _Slambda(var, lbda): #Soft thresholding Operator
    out = sp.sign(var) * sp.maximum(sp.absolute(var) - lbda/2, 0.0)
    return out


def _e_vect(n,N):
    out = np.zeros((1,N)).T
    out[n,0] = 1
    return out

def _Lmake(N, W=None):
    L = np.zeros((N,N))
    if W==None:
            L[:,:] = -1
            L += N*np.eye(N)
    else:
        for nn in range(N):
            for mm in range(nn,N):
                L += W[nn,mm]*np.dot(_e_vect(nn,N)-_e_vect(mm,N), (_e_vect(nn,N)-_e_vect(mm,N)).T)
    return L

def rmdsw(D, lbda=0.5, Ndim=3, W=None, Xinit=None, Maxit=5000, EpsLim=10**-6, EpsType="Forero2012", verbose=1):
    
    Nr = D.shape[0]
    
    if W == None: #On prend tous les éléments de D en compte: wij=1 pour tt (i,j)
        W = np.ones((Nr,Nr))
        np.fill_diagonal(W, 0)
    
    Wflat = squareform(W)    
    Dflat = squareform(D)
        
    X = np.zeros((Maxit, Ndim, Nr))
    if Xinit == None:
        X[0,:,:] = np.random.randn(Ndim, Nr)
    else:
        X[0,:,:] = Xinit.T # Xinit.T
    
    O = np.zeros((Maxit, Nr*(Nr-1)/2))
    
    L = _Lmake(Nr, W)
    Lpinv = linalg.pinv(L)
    
    A1 = np.zeros((Nr,Nr))
    A11 = np.zeros((Nr*(Nr-1)/2,))
    L1 = np.zeros((Nr,Nr))
    
    Eps = np.zeros(Maxit,)
    Err = np.zeros_like(Eps)
    
    for t in range(Maxit-1):
        if verbose: print(f't: {t:d} , Eps(t-1): { Eps[t-1]:.7f}')
        DDt = pdist(X[t,:,:].T)    
        #Calcul O(t+1):
        for nm in range(Nr*(Nr-1)//2):
                O[t+1,nm] = _Slambda(Wflat[nm]*(Dflat[nm] - DDt[nm]), lbda)
        
        #Calcul X(t+1)
        #Calcul de L1(O(t+1), X(t))
        for nm in range(Nr*(Nr-1)//2):
                if DDt[nm] != 0 and Dflat[nm] > O[t+1,nm]:
                    A11[nm] = Wflat[nm] * (Dflat[nm]-O[t+1,nm])/DDt[nm]
                else:
                    A11[nm] = 0
        
        A1 = squareform(A11)
        L1 = sp.diag(A1.sum(1)) - A1
        #Calcul X(t+1)
        X[t+1,:,:] = sp.dot(X[t,:,:], sp.dot(L1, Lpinv))
        if EpsType=="Forero2012":
            Eps[t] = linalg.norm(X[t+1,:,:]-X[t,:,:])/linalg.norm(X[t+1,:,:])
        elif EpsType=="meters":
            Eps[t] = linalg.norm(X[t+1,:,:]-X[t,:,:])
        #Err[t] = cal.errXthXest(ant.Xmic[:,0:2], X[t,:,:].T)
        if Eps[t] < EpsLim:
            break
    
    Err = Err[0:t+1]
    Eps = Eps[0:t+1]
    X = X[0:t+1,:,:].transpose(0,2,1)
    O = O[0:t+1,:]
    
    return X, O, Eps

def rmds_LambdaTune(D, lbdas, NOutlierLim, Ndim, Dlim, Xinit, Maxit, EpsLim, EpsType):
    
    Nlbda = lbdas.size
    Nr = D.shape[0]
    W = makeWlocal(D, Dlim)
    NW =  squareform(W).sum()
    
    #antExp.SetX(np.concatenate((Xsmacof, np.zeros((Nr,3-Ndim))), axis=1) )
    #print "smacof ok"
    #raw_input()
    
    Xest = np.zeros((Nlbda, Nr, Ndim))
    Oest = np.zeros((Nlbda, Nr*(Nr-1)//2))
    NOutliers = np.zeros((Nlbda,))
    
    for itlbda in range(Nlbda):
        print(f'lambda: {itlbda:d}/{Nlbda:d} ({lbdas[itlbda]:.3f})')
        lbda = lbdas[itlbda]
        
        if itlbda == 0:
            X, O, Eps = rmdsw(D, lbda, Ndim, W, Xinit, Maxit, EpsLim, EpsType, 1)
        else:
            X, O, Eps = rmdsw(D, lbda, Ndim, W, Xest[itlbda-1,:,:], Maxit, EpsLim, EpsType, 0)
        
        Oest[itlbda,:] = O[-1,:]
        Xest[itlbda,:,:] = X[-1,:,:]
        #Errest[itlbda] = Err[-1]
        NOutliers[itlbda] = np.where(O[-1,:] != 0)[0].__len__()
        
        #print '%.1f p100 d''outliers' %(100 * NOutliers[itlbda] / NW)
        #pl.pause(0.01)
        
        if NOutliers[itlbda] / NW > NOutlierLim :
            Oest = Oest[0:itlbda, :]
            Xest = Xest[0:itlbda, :, :]
            NOutliers = NOutliers[0:itlbda]
            lbdas = lbdas[0:itlbda]
            break  #!!!problème, le break n'est pas pris en compte

    return Xest, Oest, NOutliers, lbdas

###############################################################################
#Quelques ondelettes maison...:

def haar_ort(N, n):
    out = np.zeros((N,))
    out[N*(2**n - 1)/2**(n+1) : N//2] = 1.
    out[N//2 : N*(2**n + 1)/2**(n+1)] = -1.
    out *= np.sqrt(2**n/N)
    return out

def haar(N, a):
    """
    Fenêtre de Haar de N samples, égal à:
    1/(2a) entre -a et 0, -1/(2a) entre 0 et a, et 0 ailleurs.
    """
    if a > N//2: a = N//2
    out = np.zeros((N,))
    out[N//2-a : N//2] = 1.
    out[N//2 : N//2+a] = -1.
    out *= np.sqrt(1./(2*a))
    return out

def saw(N, a):
    if a>N//2: a = N//2
    out = np.zeros((N,))
    out[N//2-a : N//2] = np.linspace(0,1,a)
    out[N//2 : N//2+a] = np.linspace(-1,0,a)
    out *= np.sqrt(2**0.5/a)
    return out

###############################################################################
#Estimation D

def _fNMopt(d, mici, micj, C, freq, c0):
    """
    fNMopt(d, mici, micj, Coherence, freq)
    
    Fonction qui retourne l'erreur quadratique entre le modèle du champ 
    diffus en sinc et la cohérence mesurée.
    Pratique pour utiliser avec scipy.optimize.fmin (méthode Nelder-Mead)
    """
    return np.sum( (C[:,mici,micj].real - np.sinc(2.0*freq*d/c0))**2*sig.windows.tukey(len(freq),0.2))


import scipy.special
def _fNMoptStBern(d, mici, micj, C, freq, c0):
    """
    fNMopt(d, mici, micj, Coherence, freq)
    
    Fonction qui retourne l'erreur quadratique entre le modèle du champ 
    diffus en sinc et la cohérence mesurée.
    Pratique pour utiliser avec scipy.optimize.fmin (méthode Nelder-Mead)
    """
    return np.sum( (C[:,mici,micj].real - scipy.special.jv(0,2.0*np.pi*freq*d/c0))**2 )

def _fLMopt(d, mici, micj, C, freq, c0):
    """
    fLMopt(d, mici, micj, Coherence, freq)
    
    Fonction qui retourne la différence entre le modèle du champ 
    diffus en sinc et la cohérence mesurée.
    
    Levenburg-Marquardt
    """
    return C[:,mici,micj].real - np.sinc(2*freq*d/c0)*sig.windows.tukey(len(freq),0.2)



def C2Dfit(C, freq, D_init, c0, Err=False):
    """
    Estimateur de la matrice de distances D via un fit du modèle en sinus cardinal.
    
    Méthode de Nelder-Mead.
    """
    Nr = C.shape[1]
    Dopt = np.zeros((Nr,Nr))
    ErrOut = np.zeros_like(Dopt)
    
    print(f"Fit des sinc: {Nr:d} mics")
    if D_init.__class__ == np.ndarray:
        assert D_init.shape == C.shape[1:3], "D_init et C de tailles différentes"
        for ii in range(Nr):
            print(f"{ii:d}")
            for jj in range(ii+1,Nr):
                Dopt[ii,jj], ErrOut[ii,jj], z,z,z = opt.fmin(_fNMopt, [D_init[ii,jj]], (ii,jj, C, freq, c0), disp=0, full_output=1)
    else:
        for ii in range(Nr):
            print(f"{ii:d}")
            for jj in range(ii+1,Nr):
                Dopt[ii,jj], ErrOut[ii,jj], z,z,z = opt.fmin(_fNMopt, [D_init], (ii,jj, C, freq, c0), disp=0, full_output=1)
    
    print(" ")
    if Err==True:
        return Dopt + Dopt.T, ErrOut
    else:
        return Dopt + Dopt.T

def C2DfitStBern(C, freq, D_init, c0, Err=False):
    """
    Estimateur de la matrice de distances D via un fit du modèle en sinus cardinal.
    
    Méthode de Nelder-Mead.
    """
    Nr = C.shape[1]
    Dopt = np.zeros((Nr,Nr))
    ErrOut = np.zeros_like(Dopt)
    
    print(f"Fit des sinc: {Nr:d} mics")
    if D_init.__class__ == np.ndarray:
        assert D_init.shape == C.shape[1:3], "D_init et C de tailles différentes"
        for ii in range(Nr):
            print(f"{ii:d}")
            for jj in range(ii+1,Nr):
                Dopt[ii,jj], ErrOut[ii,jj], z,z,z = opt.fmin(_fNMoptStBern, [D_init[ii,jj]], (ii,jj, C, freq, c0), disp=0, full_output=1)
       
                
    else:
        for ii in range(Nr):
            print(f"{ii:d}")
            for jj in range(ii+1,Nr):
                Dopt[ii,jj], ErrOut[ii,jj], z,z,z = opt.fmin(_fNMoptStBern, [D_init], (ii,jj, C, freq, c0), disp=0, full_output=1)
    
    print(" ")
    if Err==True:
        return Dopt + Dopt.T, ErrOut
    else:
        return Dopt + Dopt.T


def C2DfitLSq(C, freq, D_init, c0):
    """
    Estimateur de la matrice de distances D via un fit du modèle en sinus cardinal.
    
    Méthode de Levenberg-Marquardt.
    """
    Nr = C.shape[1]
    doptAnt = np.zeros((Nr,Nr))

    if D_init.__class__ == np.ndarray:
        assert D_init.shape == C.shape[1:3], "D_init et C de tailles différentes"
        for ii in range(Nr):
            print(f"{ii:d}/{Nr:d}")
            for jj in range(ii+1,Nr):
                doptAnt[ii,jj], dump = opt.leastsq(_fLMopt, [D_init[ii,jj]], (ii,jj, C, freq, c0))
                
    else:
        for ii in range(Nr):
            print(f"{ii:d}/{Nr:d}")
            for jj in range(ii+1,Nr):
                doptAnt[ii,jj], dump = opt.leastsq(_fLMopt, [D_init], (ii,jj, C, freq, c0))
    
    return doptAnt + doptAnt.T
    
def DenoiseD(D, Ndim, epslim=0.5, itmax = 100):
    """
    DenoiseD(D, Ndim, epslim=0.5, itmax = 100)
    
    Fonction de débruitage d'une matrice de distances D.
    
    Cf. "Robust Array Calibration using Time Delays with Application to Ultrasound Tomography"
    """
    
    #Pour phi1/phi2/phi3: cf. "Robust Array Calibration using Time Delays with Application to Ultrasound Tomography"

    Q = D**2
    Q = (Q+Q.T)/2 #phi1
    ii=0
    loop = True
    
    while loop:
        ii += 1
        np.fill_diagonal(Q, 0.0) #phi2
        Q = Q.clip(0.) #phi3
        u, s, v = pl.svd(Q)
        Qplus = np.dot(u[:,0:Ndim+2], np.dot(np.diag(s[0:Ndim+2]), v[0:Ndim+2,:]) )
        eps = np.linalg.norm(Qplus-Q)
        print(eps.round(3)), 
        if eps > epslim: 
            Q = Qplus
        elif ii > itmax:
            loop = False
        else:
            loop = False
    
    print("Nb itération: {ii:d}, eps={eps:.4f}")
    return Q**0.5

###############################################################################
#Fonctions d'analyse des résultats

def errXthXest(Xth, Xest):
    """
    errXthXest(Xth, Xest):
    
    Retourne l'erreur quadratique entre Xth et Xest. Invariant selon toute 
    translation/rotation sur Xth et/ou Xmes. 
    
    Inputs
    ------
    Xth: (Nr, Ndim) numpy array
        Matrice de Nr coordonnées en dimension Ndim
    
    Xth: (Nr, Ndim) numpy array
        Matrice de Nr coordonnées en dimension Ndim
        
    Output
    -------
    Erreur quadratique (ramenée en mètre)
    """
    
    assert Xth.shape == Xest.shape, "Xth et Xest de tailles différentes"
    Nr = Xth.shape[0]
    J = np.identity(Nr) - np.ones((Nr,Nr)) / Nr
    return np.linalg.norm(  np.dot(J, np.dot(Xth, np.dot(Xth.T, J)))
                            - np.dot(J, np.dot(Xest, np.dot(Xest.T, J))), 
                            ord='fro') / Nr

        
def DthVSDmes(Dth, Dmes):
    """
    plot le nuage de points Dmes(Dth)
    """
    assert Dth.shape[0]==Dmes.shape[0], "Dth et Dmes de taille différentes"
    assert np.array_equal(Dth, Dth.T) , "Dth asymétrique."
    assert np.array_equal(Dmes, Dmes.T) , "Dmes asymétrique."
    
    Nr = Dth.shape[0]
    
    idxtriu = np.triu_indices(Nr,k=1) #indices mat. trig. sup., sans diagonale
    #Dthflt = Dth.take(idxtriu[0]).take(idxtriu[1], axis=1).flatten()
    #Dmesflt = Dmes.take(idxtriu[0]).take(idxtriu[1], axis=1).flatten()
    Dthflt = Dth[idxtriu[0], idxtriu[1]].flatten()
    Dmesflt = Dmes[idxtriu[0], idxtriu[1]].flatten()
    
    pl.figure()
    pl.plot(Dthflt, Dmesflt, '*')
    pl.plot(Dthflt, Dthflt, 'k')


def rotate2D(X, mic):
    """
    rotate2D(X, mic)
    
    Applique une rotation sur X telle que le micro mic est sur l'axe (O,x).
    """
    rmic = pl.norm(X[mic,:])
    Rot = np.array([
                    [X[mic,0]/rmic, -X[mic,1]/rmic],
                    [X[mic,1]/rmic, X[mic,0]/rmic]
                    ])
    return np.dot(X, Rot)

def c0corrector(X, Xth):
    """
    c0corrector(X, Xth)
    
    Corrige l'erreur de X du au choix a priori arbitraire de la vitesse du son.
    """
    def err_cel(c):
        return errXthXest(Xth, c*X)

    cest = opt.fmin(err_cel, [1.0])
    return cest*X, cest
    
def get_D(X):
    '''
    Retourne la matrice de distances de X.
    '''
    return squareform(pdist(X))

def rmsd(V, W):
    """ Calculate Root-mean-square deviation from two sets of vectors V and W.
    """
    D = len(V[0])
    N = len(V)
    rmsd = 0.0
    for v, w in zip(V, W):
        rmsd += sum([(v[i]-w[i])**2.0 for i in range(D)])
    return np.sqrt(rmsd/N)

def kabsch(P, Q, output=False):
    """ The Kabsch algorithm

    http://en.wikipedia.org/wiki/Kabsch_algorithm

    The algorithm starts with two sets of paired points P and Q.
    P and Q should already be centered on top of each other.

    Each vector set is represented as an NxD matrix, where D is the
    the dimension of the space.

    The algorithm works in three steps:
    - a translation of P and Q
    - the computation of a covariance matrix C
    - computation of the optimal rotation matrix U

    The optimal rotation matrix U is then used to
    rotate P unto Q so the RMSD can be caculated
    from a straight forward fashion.

    """

    # Computation of the covariance matrix
    C = np.dot(np.transpose(P), Q)

    # Computation of the optimal rotation matrix
    # This can be done using singular value decomposition (SVD)
    # Getting the sign of the det(V)*(W) to decide
    # whether we need to correct our rotation matrix to ensure a
    # right-handed coordinate system.
    # And finally calculating the optimal rotation matrix U
    # see http://en.wikipedia.org/wiki/Kabsch_algorithm
    V, S, W = np.linalg.svd(C)
    d = (np.linalg.det(V) * np.linalg.det(W)) < 0.0

    if d:
        S[-1] = -S[-1]
        V[:,-1] = -V[:,-1]

    # Create Rotation matrix U
    U = np.dot(V, W)

    # Rotate P
    P = np.dot(P, U)

    if output:
        return P, rmsd(P, Q)

    return rmsd(P, Q)



# megamicros.signal.py python software for data processing
#
# ® Copyright 2024-2025 Bimea
# Author: bruno.gas@bimea.io
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
This module provides a set of tools for signal processing and analysis.

Features:
    - Compute energy, wiener entropy, Q50 and perform signal segmentation 

Documentation:
    Full MegaMicros documentation is available at: https://readthedoc.bimea.io
"""

import numpy as np
from scipy import signal

from ..exception import MuException
from ..log import log


def wiener_entropy( signal: np.ndarray, sampling_frequency: float, frame_duration: float, hop: int|None=None ):
    """ Compute Wiener entropy for a sound signal (Wiener entropy is a measure of the signal complexity)
    The entropy is computed on consecutiv frames of the signal using the following formula:
    Wiener = exp( sum( log( e^2 ) ) / N ) / sum( e^2 ) / N

    Parameters
    ----------
    signal: np.ndarray
        The sound signal to be processed
    sampling_frequency: float
        The sampling frequency of the sound signal
    frame_duration: float
        The duration in seconds of the sliding frame used for processing the Wiener entropy
    hop: int
        The number of samples between two consecutive frames default value is equal to the frame width (not implemented yet)
    
    Returns
    -------
    wiener: np.ndarray
        The Wiener entropy of the signal computed on a sliding window
    frame_width: int
        The width of the frame used for Wiener entropy computing
    """        

    # Check if the signal is a mono signal and transform it to mono if it is not
    if len( signal.shape ) > 1:
        if signal.shape[0] == 1 and signal.shape[1] > 1:
            signal = signal[0,:]
        elif signal.shape[1] == 1 and signal.shape[0] > 1:
            signal = signal[:,0]
        else:
            raise MuException( "The signal must be a mono signal" )

    # Hop is not implemented yet
    if hop is not None:
        raise MuException( "Hop is not implemented yet for Wiener entropy computing" )

    frame_width = int( frame_duration * sampling_frequency )
    if frame_width > len( signal ):
        raise MuException( "The frame width must be smaller than the signal length" )
    
    frames_number = int( len( signal ) // frame_width )
    lost_samples_number = len( signal ) % frame_width
    signal = signal[:len(signal)-lost_samples_number]
    signal = np.reshape( signal, (frames_number, frame_width) )
    spectrum = np.fft.rfft( signal, axis=1 )
    wiener = np.zeros( frames_number )
    for i in range( frames_number ):
        e_2 = np.abs( spectrum[i,:] )**2
        le = np.log( e_2 )
        wiener[i] = np.exp( np.sum( le )/frame_width ) / np.sum( e_2 ) / frame_width

    return wiener, frame_width



def energy( signal: np.ndarray, sampling_frequency: float, frame_duration: float, hop: int|None=None ):
    """ Compute energy on a sliding window for a sound signal 
    
    Parameters
    ----------
    signal: np.ndarray
        The sound signal to be processed
    sampling_frequency: int
        The sampling frequency of the sound signal
    frame_duration: float
        The duration in seconds of the sliding frame used for processing the energy
    hop: int
        The number of samples between two consecutive frames default value is equal to the frame width
    
    Returns
    -------
    energy: np.ndarray
        The energy of the signal computed on on a sliding window
    frame_width: int
        The width of the frame used for energy computing
    """

    # Check if the signal is a mono signal and transform it to mono if it is not
    if len( signal.shape ) > 1:
        if signal.shape[0] == 1 and signal.shape[1] > 1:
            signal = signal[0,:]
        elif signal.shape[1] == 1 and signal.shape[0] > 1:
            signal = signal[:,0]
        else:
            raise MuException( "The signal must be a mono signal" )

    frame_width = int( frame_duration * sampling_frequency )
    if frame_width > len( signal ):
        raise MuException( "The frame width must be smaller than the signal length" )

    # Adjust the hop size
    if hop is None:
        hop = frame_width

    # Init an array to store the energy of each frame
    energy = np.zeros((len(signal) - frame_width) // hop + 1)

    # Calculer l'énergie pour chaque fenêtre
    for i in range( 0, len( signal ) - frame_width, hop ):
        frame = signal[i:i+frame_width]
        energy[i // hop] = np.sum( frame**2 )

    return energy, frame_width




def q50( signal: np.ndarray, sampling_frequency: int, frame_duration: float=0.1 ):
    """ Compute Q50 indice (Spectral center of gravity in the upper half of the spectrum) for a sound signal 
    
    Parameters
    ----------
    signal: np.ndarray
        The sound signal to be processed
    sampling_frequency: int
        The sampling frequency of the sound signal
    frame_duration: float
        The duration in seconds of the sliding frame used for processing the Q50 indice. Default is 0.1s
        Note that some samples may be lost at the end of the signal if the frame width is not a divisor of the signal size.
    """

    if not isinstance( signal, np.ndarray ):
        raise MuException( "The signal must be a numpy array" )
    
    # Check if the signal is a mono signal and transform it to mono if it is not
    if len( signal.shape ) > 1:
        if signal.shape[0] == 1 and signal.shape[1] > 1:
            signal = signal[0,:]
        elif signal.shape[1] == 1 and signal.shape[0] > 1:
            signal = signal[:,0]
        else:
            raise MuException( "The signal must be a mono signal" )
    
    samples_number = np.size( signal )
    frame_width = int( frame_duration * sampling_frequency )
    frames_number = int( samples_number / frame_width )
    lost_samples_number = samples_number % frame_width
    signal = signal[:samples_number - lost_samples_number]
    signal = np.reshape( signal, (frames_number, frame_width) )
    spectrum = np.fft.rfft( signal, axis=1 )
    modspec2 = np.abs( spectrum )    
    modspec2 *= modspec2   

    n_freq = np.size( modspec2,1 )
    frequencies = np.array( [i for i in range( n_freq )] ) * sampling_frequency / n_freq / 2
    
    q50 = np.zeros( frames_number )
    for i in range( frames_number ):
        e = np.abs( spectrum[i,:] )
        e_2 = np.abs( e * e )
        ew = e_2 * frequencies
        q50[i] = np.sum( ew ) / np.sum( e_2 )

    return q50, frequencies

def segment( signal: np.ndarray, seg_threshold: float=50, seg_interval: float=10, frame_width: int|None=None, frame_overlap: int|None=None):
    """ Segment a positive defined signal (energy, Q50, etc.). Segmentation is performed using a threshold and a minimum interval.

    Parameters
    ----------
    q50: np.ndarray
        The signal to be segmented
    seg_threshold: float
        The threshold used for segmentation in percent of the max amplitude. Default is 50%
    seg_interval: float
        The minimum interval for a segment in percent of the length of the signal. Default is 10%
    frame_width: int
        The width of the frame used for anterior signal generation. Default is None
    frame_overlap: int
        The overlap of the frame used for anterior signal generation. Default is None

    Returns
    -------
    signal_segments: np.ndarray
        The segmented signal as a list of (begin, end) segments tuples
    signal_seg: np.ndarray
        The segmented signal as a binary [0,1] signal of same length as the input signal
    """

    if frame_width is not None or frame_overlap is not None:
        raise MuException( "frame_width and frame_overlap are not yet implemented" )
    
    # Check if the signal is a mono signal and transform it to mono if it is not
    if len( signal.shape ) > 1:
        if signal.shape[0] == 1 and signal.shape[1] > 1:
            signal = signal[0,:]
        elif signal.shape[1] == 1 and signal.shape[0] > 1:
            signal = signal[:,0]
        else:
            raise MuException( "The signal must be a mono signal" )
    
    samples_number = np.size( signal )
    signal_max = np.max( signal )
    actual_threshold = seg_threshold * signal_max / 100
    signal_seg = np.empty( samples_number )
    signal_seg.fill( 0 )

    kernel_size = int( seg_interval * samples_number // 100 )
    i = 0
    signal_segments = []
    while i < samples_number:
        if signal[i] >= actual_threshold:
            j = i
            while j < samples_number and signal[j] >= actual_threshold:
                # browse the signal while the threshold is reached
                j += 1  
            if j - i >= kernel_size:
                # segment found
                signal_seg[i:j] = 1
                signal_segments.append( (i,j) )
            else:
                # segment too short
                pass
            i = j
        else:
            i += 1

    return signal_segments, signal_seg


# -*- coding: utf-8 -*-

"""
Module contenant toutes les fonctions pour le calcul
de covariance et cohérence sur GPU.

Created on Wed May 7 14:27:01 2014

@author: charles
"""

from pycuda.compiler import SourceModule as _SourceModule
import pycuda.gpuarray as _gpuarray
import scikits.cuda.cufft as _cufft

import numpy as _np

##############################################################################
#kernels gpu pour la covariance/cohérence

def _Compil_Pw2cov(): 
    """
    Estimateur de matrice covariance: Pw2cov(Pw,R,ksel,Nr,NFFT,B)
    """
    modPw2cov = _SourceModule("""
#include <pycuda-complex.hpp>
#include <stdio.h>
    
__global__ void Pw2cov(pycuda::complex<float> *Pw, pycuda::complex<float> *R, int *ksel, int Nr, int NFFT, int B)
{//--------------------------------------------------------------------------------
int i =  blockIdx.x*blockDim.x + threadIdx.x ; //indice suivant ligne de R
int j =  blockIdx.y*blockDim.y + threadIdx.y ; //indice suivant col de R
int k =  blockIdx.z*blockDim.z + threadIdx.z ; //indice suivant freq bin

pycuda::complex<float> Rijk (0.0,0.0);
pycuda::complex<float> normalisation(1.0/B,0.0);

int Bsize = NFFT/2 +1; // fonction rfft sur NFFT points supprime la symétrie 

int idxi;
int idxj;

for(int b=0;b<B;b++)
    {
    idxi = (i*B+b)*Bsize + ksel[k];
    idxj = (j*B+b)*Bsize + ksel[k];
    Rijk += Pw[idxi] * conj(Pw[idxj]);
    }

Rijk *= normalisation;
R[k*Nr*Nr + i*Nr + j] = Rijk;
//printf("%d %f ",k, ksel[k]);
}
""")
    return modPw2cov.get_function("Pw2cov")

def _Compil_Pw2coh():
    """
    Estimateur de matrice cohérence: Pw2coh(Pw,R,ksel,Nr,NFFT,B)
    """
    modPw2coh = _SourceModule("""
#include <pycuda-complex.hpp>
#include <stdio.h>
    
__global__ void Pw2coh(pycuda::complex<float> *Pw, pycuda::complex<float> *R, int *ksel, int Nr, int NFFT, int B)
{//--------------------------------------------------------------------------------
int i =  blockIdx.x*blockDim.x + threadIdx.x ; //indice suivant ligne de R
int j =  blockIdx.y*blockDim.y + threadIdx.y ; //indice suivant col de R
int k =  blockIdx.z*blockDim.z + threadIdx.z ; //indice suivant freq bin

pycuda::complex<float> Rijk (0.0,0.0);
pycuda::complex<float> normalisation(1.0/B,0.0);

int Bsize = NFFT/2 +1; // fonction rfft sur NFFT points supprime la symétrie 

int idxi;
int idxj;

for(int b=0;b<B;b++)
    {
    idxi = (i*B+b)*Bsize + ksel[k];
    idxj = (j*B+b)*Bsize + ksel[k];
    Rijk += Pw[idxi] * conj(Pw[idxj]) / pycuda::complex<float>(abs(Pw[idxi])*abs(Pw[idxj]),0.0) ;
    }

Rijk *= normalisation;
R[k*Nr*Nr + i*Nr + j] = Rijk;
}
""")
    return modPw2coh.get_function("Pw2coh")


##############################################################################
#kernels gpu pour la covariance/coherence récursive
def _Compil_Pw2covRec(): 
    """
    Estimateur recursif de matrice covariance: Pw2cov(Pw,R,ksel,Nr,NFFT,B)
    """
    modPw2covRec=_SourceModule("""
#include <pycuda-complex.hpp>
    
__global__ void Pw2covRec(pycuda::complex<float> *Pw, pycuda::complex<float> *R, int *ksel, int Nr, int NFFT, int b)
{//--------------------------------------------------------------------------------
int i =  blockIdx.x*blockDim.x + threadIdx.x ; //indice suivant ligne de R
int j =  blockIdx.y*blockDim.y + threadIdx.y ; //indice suivant col de R
int k =  blockIdx.z*blockDim.z + threadIdx.z ; //indice suivant freq bin

pycuda::complex<float> Rijk (0.0,0.0);

int Bsize = NFFT/2 +1; // fonction rfft sur NFFT points supprime la symétrie 
int idxi = Bsize*i+ksel[k];
int idxj = Bsize*j+ksel[k];

Rijk = Pw[idxi] * conj(Pw[idxj]);

R[k*Nr*Nr + i*Nr + j] = (R[k*Nr*Nr + i*Nr + j]*pycuda::complex<float>(b-1,0.0) + Rijk)/pycuda::complex<float>(b,0.0);
}
""")
    return modPw2covRec.get_function("Pw2covRec")
    
def _Compil_Pw2covRecDouble(): 
    """
    Estimateur recursif de matrice covariance: Pw2cov(Pw,R,ksel,Nr,NFFT,B)
    """
    modPw2covRec=_SourceModule("""
#include <pycuda-complex.hpp>
    
__global__ void Pw2covRec(pycuda::complex<double> *Pw, pycuda::complex<double> *R, int *ksel, int Nr, int NFFT, int b)
{//--------------------------------------------------------------------------------
int i =  blockIdx.x*blockDim.x + threadIdx.x ; //indice suivant ligne de R
int j =  blockIdx.y*blockDim.y + threadIdx.y ; //indice suivant col de R
int k =  blockIdx.z*blockDim.z + threadIdx.z ; //indice suivant freq bin

pycuda::complex<double> Rijk (0.0,0.0);

int Bsize = NFFT/2 +1; // fonction rfft sur NFFT points supprime la symétrie 
int idxi = Bsize*i+ksel[k];
int idxj = Bsize*j+ksel[k];

Rijk = Pw[idxi] * conj(Pw[idxj]);

R[k*Nr*Nr + i*Nr + j] = (R[k*Nr*Nr + i*Nr + j]*pycuda::complex<double>(b-1,0.0) + Rijk)/pycuda::complex<double>(b,0.0);
}
""")
    return modPw2covRec.get_function("Pw2covRec")


def _Compil_Pw2cohRec(): 
    """
    Estimateur recursif de matrice coherence: Pw2coh(Pw,R,ksel,Nr,NFFT,B)
    """
    modPw2cohRec=_SourceModule("""
#include <pycuda-complex.hpp>
    
__global__ void Pw2cohRec(pycuda::complex<float> *Pw, pycuda::complex<float> *R, int *ksel, int Nr, int NFFT, int b)
{//--------------------------------------------------------------------------------
int i =  blockIdx.x*blockDim.x + threadIdx.x ; //indice suivant ligne de R
int j =  blockIdx.y*blockDim.y + threadIdx.y ; //indice suivant col de R
int k =  blockIdx.z*blockDim.z + threadIdx.z ; //indice suivant freq bin

pycuda::complex<float> Rijk (0.0,0.0);

int Bsize = NFFT/2 +1; // fonction rfft sur NFFT points supprime la symétrie 
int idxi = Bsize*i+ksel[k];
int idxj = Bsize*j+ksel[k];

Rijk = Pw[idxi] * conj(Pw[idxj]) / pycuda::complex<float>(abs(Pw[idxi])*abs(Pw[idxj]),0.0);

R[k*Nr*Nr + i*Nr + j] = (R[k*Nr*Nr + i*Nr + j]*pycuda::complex<float>(b-1,0.0) + Rijk)/pycuda::complex<float>(b,0.0);
}
""")
    return modPw2cohRec.get_function("Pw2cohRec")
    
def _Compil_Pw2cohRecDouble(): 
    """
    Estimateur recursif de matrice coherence: Pw2coh(Pw,R,ksel,Nr,NFFT,B)
    """
    modPw2cohRec=_SourceModule("""
#include <pycuda-complex.hpp>
    
__global__ void Pw2cohRec(pycuda::complex<double> *Pw, pycuda::complex<double> *R, int *ksel, int Nr, int NFFT, int b)
{//--------------------------------------------------------------------------------
int i =  blockIdx.x*blockDim.x + threadIdx.x ; //indice suivant ligne de R
int j =  blockIdx.y*blockDim.y + threadIdx.y ; //indice suivant col de R
int k =  blockIdx.z*blockDim.z + threadIdx.z ; //indice suivant freq bin

pycuda::complex<double> Rijk (0.0,0.0);

int Bsize = NFFT/2 +1; // fonction rfft sur NFFT points supprime la symétrie 
int idxi = Bsize*i+ksel[k];
int idxj = Bsize*j+ksel[k];

Rijk = Pw[idxi] * conj(Pw[idxj]) / pycuda::complex<double>(abs(Pw[idxi])*abs(Pw[idxj]),0.0);

R[k*Nr*Nr + i*Nr + j] = (R[k*Nr*Nr + i*Nr + j]*pycuda::complex<double>(b-1,0.0) + Rijk)/pycuda::complex<double>(b,0.0);
}
""")
    return modPw2cohRec.get_function("Pw2cohRec")

_Pw2cov = _Compil_Pw2cov()
_Pw2coh = _Compil_Pw2coh()
_Pw2cohRec = _Compil_Pw2cohRec()
_Pw2cohRecDouble = _Compil_Pw2cohRecDouble()
_Pw2covRec = _Compil_Pw2covRec()
_Pw2covRecDouble = _Compil_Pw2covRecDouble()

###################################################################
# Fonctions python à appeler pour le calcul d'une:
#
# a) Covariance non récursive (DD->RAM 1x, RAM->GPU 1x)
# b) Cohérence non récursive (DD->RAM 1x, RAM->GPU 1x)
# c) Covariance récursive (DD->RAM 1x, RAM->GPU itéré par bloc)
# d) Cohérence récursive (DD->RAM 1x, RAM->GPU itéré par bloc)

def covariance(P, kSel, Nr, NFFT, B):
    """
    Estimateur de la matrice de cohérence de Nr microphones, pour un set de bins fréquentiels choisi.

    Méthode: moyennage de B produits Pi*conj(Pj), avec Pi=rfft(pi) et Pj=rfft(pj) sur NFFT points.

    Attention mémoire: transfert de l'acquisition vers GPU en 1 fois! (voir coherenceRec() sinon)
    
    Entrées
    -------
    P: (Nr, Ns) array de float32
        Acquisition des pressions des Nr microphones, Ns échantillons.
    kSel: (M,) array d'int
        Indices des bins fréquentiels où la cohérence sera calculée.
        Fréquences correspondantes à celles de la rfft de P.
    Nr: int
        nombre de microphones.
    NFFT: int
        Nombre d'échantillon des blocs pour la fft.
    B: int
       Nombre total de blocs à moyenner.
    
    Sortie
    ------
    R: (M, Nr, Nr) array de complex64
        M matrices (Nr, Nr) de cohérence aux fréquences choisies.
    """
    plan = _cufft.cufftPlan1d(NFFT, _cufft.CUFFT_R2C, Nr*B)
    Pgpu = _gpuarray.zeros((Nr,NFFT*B), _np.float32)
    Pwgpu = _gpuarray.zeros((Nr,B*(NFFT//2+1)), _np.complex64)
    Rgpu = _gpuarray.zeros((kSel.size,Nr,Nr),_np.complex64)
    kSelGPU = _gpuarray.to_gpu(kSel)

    Pgpu.set(P[:,0:NFFT*B].copy())
    _cufft.cufftExecR2C(plan,Pgpu.ptr,Pwgpu.ptr)    
    _Pw2cov(Pwgpu, Rgpu, kSelGPU, _np.int32(Nr), _np.int32(NFFT), _np.int32(B),
           grid = (1,int(Nr),int(kSel.size)), block = (int(Nr),1,1)) #x,y,z des grid/bloc <-> i,j,k
    R = Rgpu.get()

    _cufft.cufftDestroy(plan)
    for arr in [Rgpu, Pwgpu, Pgpu, kSelGPU]:
        arr.gpudata.free()    
    return R

def coherence(P, kSel, Nr, NFFT, B):
    """
    Estimateur de la matrice de cohérence de Nr microphones, pour un set de bins fréquentiels choisi.

    Méthode: moyennage de B produits Pi*conj(Pj)/(abs(Pi)*abs(Pj)), avec Pi=rfft(pi) et Pj=rfft(pj) sur NFFT points.

    Attention mémoire: transfert de l'acquisition vers GPU en 1 fois! (voir coherenceRec() sinon)
    
    Entrées
    -------
    P: (Nr, Ns) array de float32
        Acquisition des pressions des Nr microphones, Ns échantillons.
    kSel: (M,) array d'int
        Indices des bins fréquentiels où la cohérence sera calculée.
        Fréquences correspondantes à celles de la rfft de P.
    Nr: int
        nombre de microphones.
    NFFT: int
        Nombre d'échantillon des blocs pour la fft.
    B: int
       Nombre total de blocs à moyenner.
    
    Sortie
    ------
    R: (M, Nr, Nr) array de complex64
        M matrices (Nr, Nr) de cohérence aux fréquences choisies.
    """
    plan = _cufft.cufftPlan1d(NFFT, _cufft.CUFFT_R2C, Nr*B)
    Pgpu = _gpuarray.zeros((Nr,NFFT*B), _np.float32)
    Pwgpu = _gpuarray.zeros((Nr,B*(NFFT//2+1)), _np.complex64)
    Rgpu = _gpuarray.zeros((kSel.size,Nr,Nr),_np.complex64)
    kSelGPU = _gpuarray.to_gpu(kSel)

    Pgpu.set(P[:,0:NFFT*B].copy())
    _cufft.cufftExecR2C(plan,Pgpu.ptr,Pwgpu.ptr)
    _Pw2coh(Pwgpu, Rgpu, kSelGPU, _np.int32(Nr), _np.int32(NFFT), _np.int32(B),
           grid = (1,int(Nr),int(kSel.size)), block = (int(Nr),1,1)) #x,y,z des grid/bloc <-> i,j,k
    R = Rgpu.get()
    
    _cufft.cufftDestroy(plan)
    for arr in [Rgpu, Pwgpu, Pgpu, kSelGPU]:
        arr.gpudata.free()
    return R


def covarianceRec(P, kSel, Nr, NFFT, B): 
    """
    Estimateur de la matrice de covariance de Nr microphones, pour un set de bins fréquentiels choisi.

    Méthode: moyennage de B produits Pi*conj(Pj), avec Pi=rfft(pi) et Pj=rfft(pj) sur NFFT points.

    Transfert de l'acquisition vers GPU en B fois, mais HD->RAM  en 1 fois!
    
    Entrées
    -------
    P: (Nr, Ns) array de float32
        Acquisition des pressions des Nr microphones, Ns échantillons.
    kSel: (M,) array d'int
        Indices des bins fréquentiels où la cohérence sera calculée.
        Fréquences correspondantes à celles de la rfft de P.
    Nr: int
        nombre de microphones.
    NFFT: int
        Nombre d'échantillon des blocs pour la fft.
    B: int
       Nombre total de blocs à moyenner.
    
    Sortie
    ------
    R: (M, Nr, Nr) array de complex64
        M matrices de cohérence aux fréquences choisies.
    """    
    plan = _cufft.cufftPlan1d(NFFT, _cufft.CUFFT_R2C, Nr)
    Pgpu = _gpuarray.zeros((Nr,NFFT), _np.float32)
    Pwgpu = _gpuarray.zeros((Nr,(NFFT//2+1)), _np.complex64)
    Rgpu = _gpuarray.zeros((kSel.size,Nr,Nr),_np.complex64)
    kSelGPU = _gpuarray.to_gpu(kSel)
    
    for b in range(1,B+1):
        Pbuf = P[:,(b-1)*NFFT:b*NFFT].copy()
        Pgpu.set(Pbuf)
        _cufft.cufftExecR2C(plan,Pgpu.ptr,Pwgpu.ptr)
        _Pw2covRec(Pwgpu, Rgpu, kSelGPU, _np.int32(Nr), _np.int32(NFFT), _np.int32(b),
           grid = (1,int(Nr),int(kSel.size)), block = (int(Nr),1,1)) #x,y,z des grid/bloc <-> i,j,k
    
    Rrec = Rgpu.get()

    for arr in [Rgpu, Pwgpu, Pgpu, kSelGPU]:
        arr.gpudata.free()
    return Rrec
    
def covarianceRecdouble(P, kSel, Nr, NFFT, B): 
    """
    Estimateur de la matrice de covariance de Nr microphones, pour un set de bins fréquentiels choisi.
    
    !Double Précision!

    Méthode: moyennage de B produits Pi*conj(Pj), avec Pi=rfft(pi) et Pj=rfft(pj) sur NFFT points.

    Transfert de l'acquisition vers GPU en B fois, mais HD->RAM  en 1 fois!
    
    Entrées
    -------
    P: (Nr, Ns) array de float64
        Acquisition des pressions des Nr microphones, Ns échantillons.
    kSel: (M,) array d'int
        Indices des bins fréquentiels où la cohérence sera calculée.
        Fréquences correspondantes à celles de la rfft de P.
    Nr: int
        nombre de microphones.
    NFFT: int
        Nombre d'échantillon des blocs pour la fft.
    B: int
       Nombre total de blocs à moyenner.
    
    Sortie
    ------
    R: (M, Nr, Nr) array de complex128
        M matrices de cohérence aux fréquences choisies.
    """    
    plan = _cufft.cufftPlan1d(NFFT, _cufft.CUFFT_D2Z, Nr)
    Pgpu = _gpuarray.zeros((Nr,NFFT), _np.float64)
    Pwgpu = _gpuarray.zeros((Nr,(NFFT//2+1)), _np.complex128)
    Rgpu = _gpuarray.zeros((kSel.size,Nr,Nr),_np.complex128)
    kSelGPU = _gpuarray.to_gpu(kSel)
    
    for b in range(1,B+1):
        Pbuf = P[:,(b-1)*NFFT:b*NFFT].astype(_np.float64).copy()
        Pgpu.set(Pbuf)
        _cufft.cufftExecD2Z(plan,Pgpu.ptr,Pwgpu.ptr)
        _Pw2covRecDouble(Pwgpu, Rgpu, kSelGPU, _np.int32(Nr), _np.int32(NFFT), _np.int32(b),
           grid = (1,int(Nr),int(kSel.size)), block = (int(Nr),1,1)) #x,y,z des grid/bloc <-> i,j,k
    
    Rrec = Rgpu.get()

    for arr in [Rgpu, Pwgpu, Pgpu, kSelGPU]:
        arr.gpudata.free()
    return Rrec

def coherenceRec(P, kSel, Nr, NFFT, B):
    """
    Estimateur de la matrice de cohérence de Nr microphones, pour un set de bins fréquentiels choisi.

    Méthode: moyennage de B produits Pi*conj(Pj)/(abs(Pi)*abs(Pj)), avec Pi=rfft(pi) et Pj=rfft(pj) sur NFFT points.

    Transfert de l'acquisition vers GPU en B fois, mais HD->RAM  en 1 fois!
    
    Entrées
    -------
    P: (Nr, Ns) array de float32
        Acquisition des pressions des Nr microphones, Ns échantillons.
    kSel: (M,) array d'int
        Indices des bins fréquentiels où la cohérence sera calculée.
        Fréquences correspondantes à celles de la rfft de P.
    Nr: int
        nombre de microphones.
    NFFT: int
        Nombre d'échantillon des blocs pour la fft.
    B: int
       Nombre total de blocs à moyenner.
    
    Sortie  
    ------
    R: (M, Nr, Nr) array de complex64
        M matrices de cohérence aux fréquences choisies.
    """
    plan = _cufft.cufftPlan1d(NFFT, _cufft.CUFFT_R2C, Nr)
    Pgpu = _gpuarray.zeros((Nr,NFFT), _np.float32)
    Pwgpu = _gpuarray.zeros((Nr,(NFFT//2+1)), _np.complex64)
    Rgpu = _gpuarray.zeros((kSel.size,Nr,Nr),_np.complex64)
    kSelGPU = _gpuarray.to_gpu(kSel)
    
    for b in range(1,B+1):
        Pbuf = P[:,(b-1)*NFFT:b*NFFT].copy()
        Pgpu.set(Pbuf)
        _cufft.cufftExecR2C(plan,Pgpu.ptr,Pwgpu.ptr)
        _Pw2cohRec(Pwgpu, Rgpu, kSelGPU, _np.int32(Nr), _np.int32(NFFT), _np.int32(b),
           grid = (1,int(Nr),int(kSel.size)), block = (int(Nr),1,1)) #x,y,z des grid/bloc <-> i,j,k
    
    Rrec = Rgpu.get()
    
    _cufft.cufftDestroy(plan)
    for arr in [Rgpu, Pwgpu, Pgpu, kSelGPU]:
        arr.gpudata.free()
    return Rrec


def coherenceRecdouble(P, kSel, Nr, NFFT, B):
    """
    Estimateur de la matrice de cohérence de Nr microphones, pour un set de bins fréquentiels choisi.
    
    !Double Précision!
    
    Méthode: moyennage de B produits Pi*conj(Pj)/(abs(Pi)*abs(Pj)), avec Pi=rfft(pi) et Pj=rfft(pj) sur NFFT points.

    Transfert de l'acquisition vers GPU en B fois, mais HD->RAM  en 1 fois!
    
    Entrées
    -------
    P: (Nr, Ns) array de float64
        Acquisition des pressions des Nr microphones, Ns échantillons.
    kSel: (M,) array d'int
        Indices des bins fréquentiels où la cohérence sera calculée.
        Fréquences correspondantes à celles de la rfft de P.
    Nr: int
        nombre de microphones.
    NFFT: int
        Nombre d'échantillon des blocs pour la fft.
    B: int
       Nombre total de blocs à moyenner.
    
    Sortie
    ------
    R: (M, Nr, Nr) array de complex128
        M matrices de cohérence aux fréquences choisies.
    """
    plan = _cufft.cufftPlan1d(NFFT, _cufft.CUFFT_D2Z, Nr)
    Pgpu = _gpuarray.zeros((Nr,NFFT), _np.float64)
    Pwgpu = _gpuarray.zeros((Nr,(NFFT//2+1)), _np.complex128)
    Rgpu = _gpuarray.zeros((kSel.size,Nr,Nr),_np.complex128)
    kSelGPU = _gpuarray.to_gpu(kSel)
    
    for b in range(1,B+1):
        Pbuf = P[:,(b-1)*NFFT:b*NFFT].astype(_np.float64).copy()
        Pgpu.set(Pbuf)
        _cufft.cufftExecD2Z(plan,Pgpu.ptr,Pwgpu.ptr)
        _Pw2cohRecDouble(Pwgpu, Rgpu, kSelGPU, _np.int32(Nr), _np.int32(NFFT), _np.int32(b),
           grid = (1,int(Nr),int(kSel.size)), block = (int(Nr),1,1)) #x,y,z des grid/bloc <-> i,j,k
    
    Rrec = Rgpu.get()
    
    _cufft.cufftDestroy(plan)
    for arr in [Rgpu, Pwgpu, Pgpu, kSelGPU]:
        arr.gpudata.free()
    return Rrec



# megamicros.core.base.py
#
# ® Copyright 2024-2025 Bimea
# Author: bruno.gas@bimea.io
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
This module defines the base class for all microphone arrays in the Megamicros library.

Features:
    - Management of available and active microphones
    - Configuration of sampling frequency, data type, frame length, and acquisition duration
    - Support for counter activation and skipping in output streams
    - H5 local recording flag

Examples:
    Basic usage::

        from megamicros import log
        from megamicros.base import MemsArray

        antenna = MemsArray()
        antenna.run(
            mems=(0,1,2,3),
            sampling_frequency=44100,
            duration=60,
            frame_length=1024,
            datatype='int32'
        )
        antenna.wait()

    Advanced usage::

        See the Notebooks for advanced usage examples.

Note:
    The MemsArray is not fully implemented and acts as an abstract class for specific array implementations.

See Also:
    megamicros.mu for Megamicros class implementation.

Documentation:
    Full MegaMicros documentation is available at: https://readthedoc.bimea.io
"""

import time
import queue
import numpy as np
from threading import Thread, Timer

from ..log import log
from ..exception import MuException

TRANSFER_DATAWORDS_SIZE		    = 4 									    # Size of transfer words in bytes (same for in32 and float32 data type which always states for 32 bits (-> 4 bytes) )

DEFAULT_FRAME_LENGTH            = 1024                                      # Default frame length in samples number for data transfer
DEFAULT_ACQ_DURATION            = 0                                         # Default acquisition duration in seconds (0 = infinite loop)
DEFAULT_SAMPLING_FREQUENCY      = 44100                                     # Default system sample rate for audio acquisition
DEFAULT_DATATYPE                = 'int32'                                   # Default datatype (int32 or float32)
DEFAULT_MEMS_SENSIBILITY        = 3.54e-6                                   # Default MEMS sensitivity (racine(2)/400 000 = 3,54µPa/digit)
DEFAULT_QUEUE_SIZE              = 0                                         # Queue limit after which the last signal is lost (0 means infinite signal queueing)
DEFAULT_QUEUE_TIMEOUT           = 1000                                      # Queue timeout in ms
DEFAULT_AVAILABLE_MEMS          = [i for i in range(32)]                    # Default available MEMs list for Megamicros systems (32 MEMs)
 

class MemsArray :
    """
    MemsArray
    
    Base class to handle Antenna devices
    """

    class Queue( queue.Queue ):
        """ Thread safe embedded queue adapted to the MemsArray class
        
        Original comportment is that of blocks insertion once maxsize is reached, until queue items are consumed.
        This implementation pop up the last element of the queue if maxsize is reached.

        Queue elements are binary data blocks as received from the data generator.
        """
    
        def __init__( self, maxsize: int = 0 ):
            self.__transfert_lost: int = 0
            super().__init__( maxsize )

        @property
        def transfert_lost( self ) -> bool:
            """ Get the transfer lost counter """
            return self.__transfert_lost

        @property
        def queue_size( self ) -> int:
            """ Get the queue size (e.g. maxsize) """
            return self.maxsize

        def clear( self ) -> None:
            """ Clear the queue """
            # Manually empty the queue since Queue doesn't have a clear() method
            while not self.empty():
                try:
                    self.get_nowait()
                except queue.Empty:
                    break
            self.__transfert_lost = 0

        def put( self, data ):
            """ Put data in the queue. Pop up the last element if maxsize is reached """
            if self.maxsize > 0 and self.qsize() >= self.maxsize:
                # maxsize is reached: remove the last element to free space
                self.get()
                self.__transfert_lost += 1

            # Let parent class do the work 
            super().put( data )

    def __init__(self):
        """
        @brief Constructor
        """
        self.__available_mems: list[int]=DEFAULT_AVAILABLE_MEMS       # Available microphones (connected and ok on the antenna)
        self.__mems: list[int]=[]                                     # Activated microphones
        self.__mems_sensibility: float=DEFAULT_MEMS_SENSIBILITY       # MEMS sensibility in Pa/digit (default to 3.54e-6 Pa/digit)
        self.__mems_positions: list[list[float]]=[]                   # Microphones positions vectors
        self.__counter: bool=False                                    # Counter activation flag
        self.__counter_skip: bool=False                               # Whether to skip counter in output data stream
        self.__sampling_frequency: int=DEFAULT_SAMPLING_FREQUENCY     # Default system sample rate for audio acquisition
        self.__datatype: str=DEFAULT_DATATYPE                         # "int32" or "float32"
        self.__duration: int=DEFAULT_ACQ_DURATION                     # acquisition duration in seconds
        self.__frame_length: int=DEFAULT_FRAME_LENGTH                 # Frame length in samples number for data transfer
        self.__it = 0                                                 # Iterator index
        self.__running: bool = False                                  # Running flag
        self.__source_position: list[float] = [0.0, 0.0, np.inf]      # Source position in space [x,y,z] (default to far field)
        self.__halt_request: bool = False                             # Halt request flag True if a running stop has been requested
        self.__h5_recording: bool=False                               # H5 local recording flag

        # Queue management
        self.__queue: MemsArray.Queue = MemsArray.Queue( maxsize=DEFAULT_QUEUE_SIZE )

        # Thread management
        self.__timer_thread: Thread|None = None                         # Timer thread for limited duration runs
        self.__async_transfer_thread: Thread|None = None                # Asynchronous transfer thread
        self.__async_transfer_thread_exception: Exception|None = None   # Exception raised in the async thread

    @property
    def sampling_frequency( self ) -> int:
        return self.__sampling_frequency

    @property
    def mems( self ) -> list[int]:
        return self.__mems

    @property
    def available_mems( self ) -> list[int]:
        return self.__available_mems

    @property
    def mems_sensibility( self ) -> float:
        return self.__mems_sensibility

    @property
    def mems_position( self ) -> list[list[float]]:
        return self.__mems_positions

    @property
    def counter( self ) -> bool:
        return self.__counter
    
    @property
    def counter_skip( self ) -> bool:
        return self.__counter_skip

    @property
    def channels_number( self ) -> int:
        return len(self.mems) + (1 if self.counter else 0)

    @property
    def duration( self ) -> int:
        return self.__duration
    
    @property
    def datatype( self ) -> str:
        return self.__datatype
    
    @property
    def frame_length( self ) -> int:
        return self.__frame_length
    
    @property
    def frame_duration( self ) -> float:
        return self.__frame_length / self.sampling_frequency
    
    @property
    def h5_recording( self ) -> bool:
        return self.__h5_recording

    @property
    def queue( self ) -> 'MemsArray.Queue':
        """ Get the data transfer queue """
        return self.__queue

    @property
    def queue_size( self ) -> int:
        """ Get the queue size (e.g. maxsize) in bytes"""
        return self.__queue.maxsize

    @property
    def queue_length( self ) -> int:
        """ Get the queue length in number of frames """
        if self.channels_number == 0 or self.frame_length == 0:
            return 0
        else:
            return self.__queue.maxsize // ( self.frame_length * self.channels_number * TRANSFER_DATAWORDS_SIZE )

    @property
    def queue_content( self ) -> int:
        """ Get the number of elements currently in the queue """
        return self.__queue.qsize()

    @property
    def transfer_lost( self ) -> int:
        """ Get the transfer lost counter """
        return self.__queue.transfert_lost
    
    @property
    def running( self ) -> bool:
        """ Get the running flag. True if a run thread is active, False otherwise """
        return self.__running
    
    @property
    def source_position( self ) -> list[float]:
        """ Get source position in space [x,y,z] """
        return self.__source_position

    @property
    def infos( self ) -> dict :
        """ Get current MemsArray configuration as a dictionary """
        infos_dict = {
            'available_mems': self.__available_mems,
            'active_mems': self.__mems,
            'sampling_frequency': self.__sampling_frequency,
            'datatype': self.__datatype,
            'duration': self.__duration,
            'frame_length': self.__frame_length,
            'frame_duration': self.frame_duration,
            'channels_number': self.channels_number,
            'counter': self.__counter,
            'mems_sensibility': self.__mems_sensibility,
            'mems_positions': self.__mems_positions,
            'h5_recording': self.__h5_recording
        }
        return infos_dict

    def setQueueLength( self, length: int ) -> None:
        """ Set transfer queue length in number of frames

        Parameters
        ----------
        length : int
            The transfer queue length in number of frames
        """
        self.__queue = MemsArray.Queue( maxsize=length * self.frame_length * self.channels_number * TRANSFER_DATAWORDS_SIZE )

    def setDuration( self, duration: int ) -> None:
        """ Set duration of next acquisition run in seconds 

        Parameters
        ----------
        duration: int
            The acquisition duration in seconds (0 = infinite loop)
        """
        self.__duration = duration

    def setAvailableMems( self, mems: list[int] ) -> None:
        """ Set the available MEMs (connected and ok on the antenna)

        Parameters
        ----------
        mems: list[int]
            The available MEMs list
        """
        self.__available_mems = mems

    def setActiveMems( self, mems: tuple ) -> None :
        """ Activate mems
        
        Parameters:
        -----------
        mems : tuple
            list or tuple of mems number to activate
        """

        # Set parent property
        self.__mems = mems

    def setSamplingFrequency( self, sampling_frequency: int ) -> None:
        self.__sampling_frequency = sampling_frequency

    def setCounter( self, counter: bool=True ) -> None:
        self.__counter = counter

    def setCounterSkip( self, counter_skip: bool=True ) -> None:
        """ Set whether to skip counter in output data stream
        
        Parameters:
        -----------
        counter_skip : bool
            Whether to skip counter in output data stream
        """

        if counter_skip and not self.counter:
            log.warning( " .Counter skip requested but counter is not activated" )

        self.__counter_skip = counter_skip

    def setFrameLength( self, length: int ) -> None:
        """ Set the frame length in samples number. This property also updates the USB buffer length in samples number.

        Parameters
        ----------
        length: int
            The frame length / USB buffer length in samples number
        """
        self.__frame_length = length
        self.__frame_duration = length / self.sampling_frequency

    def setDatatype( self, datatype: str ) -> None:
        """ Set data type for acquisition
        
        Parameters:
        -----------
        datatype : str
            datatype string ('int32' or 'float32')
        """

        if datatype not in ['int32', 'float32']:
            raise MuException( f"Datatype {datatype} not supported. Available datatypes are 'int32' and 'float32'" )

        self.__datatype = datatype

    def setMemsSensibility( self, sensibility: float ) -> None:
        """ Set MEMs sensibility
        
        Parameters:
        -----------
        sensibility : float
            The MEMs sensibility in Pa/digit
        """

        self.__mems_sensibility = sensibility

    def setMemsPosition( self, positions: list[list[float]] ) -> None:
        """ Set MEMs positions
        
        Parameters:
        -----------
        positions : list[list[float]]
            list of MEMs positions vectors [[x1,y1,z1], [x2,y2,z2], ...]

        Note
        ----
        The available MEMs are automatically updated to match the number of provided positions.
        """

        self.__mems_positions = positions
        self.__available_mems = list(range(len(positions)))

    def setSourcePosition( self, position: list[float] ) -> None:
        """ Set source position in space [x,y,z]
        
        Parameters:
        -----------
        position : list[float]
            The source position in space [x,y,z]
        """

        if len(position) != 3:
            raise MuException( f"Source position must be a list of 3 floats [x,y,z], got {position}" )

        self.__source_position = position

        
    def _set_run_settings( self, args, kwargs ) -> None :
        """ Set settings for run method
        
        Parameters
        ----------
        args: array
            direct arguments of the run function
        kwargs: array
            named arguments of the run function
        """
        
        if len( args ) > 0:
            log.warning( f" .Direct arguments are not accepted. Use named arguments instead ({args})" )
            raise MuException( "Direct arguments are not accepted" )

        try:
            if 'mems' in kwargs:
                self.setActiveMems( kwargs['mems'] )

            if 'available_mems' in kwargs:
                self.setAvailableMems( kwargs['available_mems'] )
                del kwargs['available_mems']

            if 'counter' in kwargs:
                self.setCounter( kwargs['counter'] )

            if 'counter_skip' in kwargs:
                self.setCounterSkip( kwargs['counter_skip'] )

            if 'sampling_frequency' in kwargs:
                self.setSamplingFrequency( kwargs['sampling_frequency'] )

            if 'mems_position' in kwargs:
                self.setMemsPosition( kwargs['mems_position'] )

            if 'datatype' in kwargs:
                self.setDatatype( kwargs['datatype'] )

            if 'duration' in kwargs:
                self.setDuration( kwargs['duration'] )

            if 'frame_length' in kwargs:
                self.setFrameLength( kwargs['frame_length'] )

        except Exception as e:
            raise MuException( f"Run failed on settings: {e}")

    def run( self, *args, **kwargs ) :
        """ The main run method that runs the antenna """

        if len( args ) > 0:
            raise MuException( f"Run() method does not accept direct arguments" )

        self._set_run_settings( [], kwargs=kwargs )

        if self.channels_number == 0 :
            raise MuException("No channel activated for data acquisition")

        # verbose
        log.info( f" .Starting run execution on Virtual device..." )
        log.info( f"  > Run infinite loop (duration=0)" if self.duration == 0 else f"  > Perform {self.duration}s run loop" )
        log.info( f"  > Sampling frequency: {self.sampling_frequency} Hz" )
        log.info( f"  > {len(self.mems)} activated microphones" )
        log.info( f"  > Activated microphones: {self.mems}" )
        log.info( f"  > MEMs sensibility: {self.mems_sensibility}" )
        log.info( f"  > Whether counter is activated: {'YES' if self.counter else 'NO'}" )
        log.info( f"  > Total channels number is {self.channels_number}" )
        log.info( f"  > Datatype: {str( self.datatype )}" )
        log.info( f"  > Frame length in samples number: {self.frame_length} samples" )
        log.info( f"  > Frame duration: {self.frame_duration} s ({self.frame_duration * 1000} ms)" )
        log.info( f"  > Frame size in bytes: {self.frame_length * self.channels_number * TRANSFER_DATAWORDS_SIZE}" )
        log.info( f"  > Transfer queue length: {'infinite queuing' if self.queue_length == 0 else self.queue_length} (frames)" )
        log.info( f"  > Transfer queue size: {'infinite queuing' if self.queue_size == 0 else self.queue_size} (bytes)" )
        log.info( f"  > Local H5 recording {'on' if self.h5_recording else 'off'}" )

        # init running flags and queue
        self.__halt_request = False
        self.__queue.clear()

        # Start the timer if a limited execution time is requested 
        if self.duration > 0:
            self.__timer_thread = Timer( self.duration, self.__run_time_ending )
            self.__timer_thread.start()

        # Start run thread
        self.__async_transfer_thread = Thread( target= self.__run_thread )
        self.__async_transfer_thread.start()


    def __run_time_ending( self ) -> None:
        """ End timer callback for run stopping. This callback requests for stopping the run thread."""

        log.info( f" .End of timer: requesting for stopping run thread..." )
        self.__halt_request = True


    def __run_thread( self ) -> None :
        """ Default base run thread execution

        Generates random data
        """

        try:
            log.info( " .Run thread execution started" )
            
            self.__running = True
            mems_number = len( self.mems )
            start_time = time.time()
            frame_duration = self.frame_duration
            frame_count = 0
            while self.__running:
                # generates random data
                if self.counter is None :
                    # send data without counter state
                    data = ( np.random.rand( self.frame_length, mems_number ) * 2 - 1 ).astype( np.float32 )
                else:
                    # add counter values
                    counter = np.array( [[i for i in range(self.frame_length)]] ).T + self.__it * self.frame_length
                    data = ( np.concatenate( ( counter, ( np.random.rand( self.frame_length, mems_number ) * 2 - 1 ) ), axis=1 ) ).astype( np.float32 )

                # post them in the internal queue as binary buffer
                self.queue.put( data.tobytes() )

                # Check for halt request
                if self.__halt_request:
                    log.info( " .Halt request received: stopping run thread..." )
                    self.__running = False

                # wait for next frame time to sensure real time generation
                frame_count += 1
                elapsed_time = time.time() - start_time
                expected_time = frame_count * frame_duration
                time_to_wait = expected_time - elapsed_time
                if time_to_wait > 0:
                    time.sleep( time_to_wait )
                    
            log.info( " .Running stopped: normal thread termination" )

        except Exception as e:
            log.error( f" .Error resulting in thread termination ({type(e).__name__}): {e}" )
            self.__async_transfer_thread_exception = e

    def wait( self ) -> None :
        """ Wait for the end of the thread execution. This is a blocking function."""
        log.info( f" .Waiting for the end of the thread execution..." )

        if self.__async_transfer_thread is not None:
            self.__async_transfer_thread.join()
            self.__async_transfer_thread = None

        if self.__async_transfer_thread_exception is not None:
            thread_exception = MuException( f"Thread exception ({type(self.__async_transfer_thread_exception).__name__}): {self.__async_transfer_thread_exception}" )
            self.__async_transfer_thread_exception = None
            raise thread_exception

    def stop( self ) -> None :
        """ Stop current running """

        log.info( " .Request for stopping thread execution" )

        if self.running:
            log.info( " .Stopping current run thread execution..." )
            self.__halt_request = True
        else:
            log.warning( "Failed to stop: No current thread running" )



    def __iter__( self ) :
        """ Init iterations over the antenna data """

        self.__it = 0
        return self

    def __next__( self ) -> np.ndarray|StopIteration :
        """ next iteration over the antenna data 
        """
        
        try:
            # Get data from the queue and format them according to datatype
            data = self.__queue.get( timeout=DEFAULT_QUEUE_TIMEOUT / 1000 )
            if self.datatype == 'int32':
                data = np.frombuffer( data, dtype=np.int32 )
                data = data.reshape( ( -1, self.frame_length ), order='F' )
            elif self.datatype == 'float32':
                data = np.frombuffer( data, dtype=np.float32 )
                data = data.reshape( ( -1, self.frame_length ), order='F' )
            else:
                raise MuException( f"Unknown datatype {self.datatype}" )
            self.__it += 1
            return data

        except queue.Empty:
            raise StopIteration

    def __len__( self) -> int : 
        """ Get the number of elements currently in the queue """

        if self.queue is None:
            return 0
        else:
            return self.queue_content



# megamicros.version.py is a simple program that prints the welcome message and the version of the program.
#
# ® Copyright 2024-2025 Bimea
# Author: bruno.gas@bimea.io
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from megamicros import __version__
from megamicros import welcome_msg

def main():
    print( welcome_msg )
    print( f"megamicros {__version__}" )


if __name__ == "__main__":
	main()



# megamicros.muh5.py H5 file handler for MegaMicros libraries
#
# Copyright (c) 2023 Sorbonne Université
# Author: bruno.gas@sorbonne-universite.fr
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Megamicros documentation is available on https://readthedoc.bimea.io
"""

import numpy as np
import h5py

from .log import log, logging
from .core.mu import MU_MEMS_SENSIBILITY

class MuH5:
    
    @property
    def sampling_frequency( self ):
        return self.__sampling_frequency

    @property
    def mems_number( self ):
        return self.__mems_number
    
    @property
    def available_mems( self ):
        return self.__available_mems

    @property
    def channels_number( self ):
        return self.__channels_number

    @property
    def duration( self ):
        return self.__duration

    @property
    def status( self ) -> bool:
        return self.__status
    
    @property
    def samples_number( self ):
        return self.__samples_number
    
    @property
    def info( self ):
        return self.__info
    
    @property
    def video_available( self ) -> bool:
        return self.__video_available
    
    @property 
    def info_video( self ):
        return self.__info_video
    
    @property
    def video_adaptive_fps( self ) -> bool:
        return self.__adaptive_fps
    
    @property
    def video_max_fps( self ) -> float:
        return self.__max_fps
    
    @property
    def video_dataset_count( self ) -> int:
        return self.__video_dataset_count

    @property
    def video_frame_count( self ) -> int:
        return self.__video_frame_count
    

    def __init__( self, filename:str ):

        """ open h5 file and get informations from """
        self.__filename = filename
        with h5py.File( filename, 'r' ) as f:

            """ Control whether H5 file is a MuH5 file """
            if not f['muh5']:
                raise Exception( f"{filename} seems not to be a MuH5 file: unrecognized format" )

            """ get fil informations """
            group = f['muh5']
            self.__info = dict( zip( group.attrs.keys(), group.attrs.values() ) )
            self.__sampling_frequency = self.__info['sampling_frequency']
            self.__available_mems = list( self.__info['mems'] )
            self.__mems_number = len( self.__available_mems )
            self.__available_analogs = list( self.__info['analogs'] )
            self.__analogs_number = len( self.__available_analogs )
            self.__duration = self.__info['duration']
            self.__counter = self.__info['counter'] and not self.__info['counter_skip']
            self.__status = True if 'status' in self.__info and self.__info['status'] else False
            self.__channels_number = self.__mems_number + self.__analogs_number + ( 1 if self.__counter else 0 ) + ( 1 if self.__status else 0 )
            self.__dataset_length = self.__info['dataset_length']
            self.__dataset_number = self.__info['dataset_number']
            self.__samples_number = self.__dataset_number * self.__dataset_length

            # Explore the first dataset to check data integrity and get data type
            dataset = f['muh5/0/sig']
            data = np.array( dataset[:] )
            if data.shape[0] != self.__channels_number or data.shape[1] != self.__dataset_length:
                raise Exception( f"Data integrity error: dataset shape {data.shape} does not match expected shape ({self.__channels_number}, {self.__dataset_length})" )

            # Check if muh5/video group exists
            self.__video_available = False
            if 'video' not in f['muh5']:
                self.__video_available = False
            else:
                self.__video_available = True

                # get video group attributes:
                video_group = f['muh5/video']
                self.__info_video = dict( zip( video_group.attrs.keys(), video_group.attrs.values() ) )
                self.__adaptive_fps = self.__info_video['adaptive_fps']
                self.__max_fps = self.__info_video['max_fps']
                self.__video_dataset_count = self.__info_video['video_dataset_count']
                self.__video_frame_count = self.__info_video['video_frame_count']

                # Explore the first dataset to check video data integrity
                # Check video dataset existance
                if '0' not in f['muh5/video']:
                    raise Exception( f"Video data integrity error: video dataset 0 not found in muh5/video group " )
                else:
                    log.info( f" .This muh5 file contains a video: {self.__video_frame_count} frames found" )

            log.info( f" .Created MuH5 object from {filename} file " )

    def get_video_frames( self, start_frame: int=0, end_frame: int=-1 ) -> np.ndarray:
        """
        Extract video from file

        Returns
        -------
        * video (np.ndarray): array of shape ( frame_count, height, width, channels ) containing video frames
        """

        if not self.__video_available:
            raise Exception( "No video available in this MuH5 file" )

        if end_frame == -1:
            end_frame = self.__video_frame_count - 1

        if start_frame < 0 or end_frame >= self.__video_frame_count or start_frame > end_frame:
            raise Exception( f"Frame index out of bounds: start_frame={start_frame}, end_frame={end_frame}, video_frame_count={self.__video_frame_count}" )

        video_frames = []

        with h5py.File( self.__filename, 'r' ) as f:
            # Get frames from start_frame to end_frame
            frame_index = 0
            for dataset_index in range( self.__video_dataset_count ):
                dataset = f['muh5/video/' + str( dataset_index ) + '/img']
                dataset_frame_number = dataset.shape[0]
                if frame_index + dataset_frame_number - 1 < start_frame:
                    # Skip this dataset
                    frame_index += dataset_frame_number
                    continue
                if frame_index > end_frame:
                    # All requested frames have been extracted
                    break
                # Extract frames from this dataset
                data = np.array( dataset[:] )
                
                # Get the range of frames to extract from this dataset
                dataset_start_frame = max( 0, start_frame - frame_index )
                dataset_end_frame = min( dataset_frame_number - 1, end_frame - frame_index )

                for index in range( dataset_start_frame, dataset_end_frame + 1 ):
                    video_frames.append( data[index,:,:,:] )
                    frame_index += 1

        video_array = np.array( video_frames )

        return video_array

    def get_signal( self, channels: list, mems_sensibility:float=MU_MEMS_SENSIBILITY ) -> np.ndarray:
        """
        Extract signal from file

        Parameters
        ----------
        * channels (list<int>): list of channels to extract
        * mems_sensibility: mems semsibility factor. if 0, the original signal is returned as it is (int32), otherwise as float32
        """

        """ build the mask from the channels list given as argument """
        mask = [ True if channel in channels else False for channel in range(self.channels_number) ]

        sound = np.zeros( ( len( channels ), self.__samples_number ), dtype=np.int32 )

        with h5py.File( self.__filename, 'r' ) as f:
            offset = 0
            for dataset_index in range( self.__dataset_number ):
                dataset = f['muh5/' + str( dataset_index ) + '/sig']
                sound[:,offset:offset+self.__dataset_length] = np.array( dataset[:] )[mask,:]
                offset += self.__dataset_length

        """ product with mems sensibility factor if one is given """
        if mems_sensibility is not None or mems_sensibility != 0:
            first_mems = 1 if self.__counter else 0
            last_mems = first_mems + self.__mems_number - 1
            mask = [ True if channel >= first_mems and channel <= last_mems else False for channel in channels ]
            sound = sound.astype( np.float32 )
            sound[mask,:] = sound[mask,:] * mems_sensibility

        return sound

    def get_one_channel_signal( self, channel_number, mems_sensibility:float=MU_MEMS_SENSIBILITY  ) -> np.ndarray:
        """
        Get only one channel signal whose channel number is given as argument
        """
        if channel_number >= self.channels_number:
            raise Exception( f"Index overflow: cannt extract channel [{channel_number}] from MuH5 with only [{self.channels_number}] channels " )

        return self.get_signal( [channel_number], mems_sensibility=mems_sensibility )

    def get_all_channels_signal( self, mems_sensibility:float=MU_MEMS_SENSIBILITY  ) -> np.ndarray:
        """
        Get all channels signal
        """        

        return self.get_signal( [i for i in range( self.__channels_number )], mems_sensibility=mems_sensibility )